#include <vector>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#include "cache.h"
#include "site.h"

unsigned long Cache::CacheNameCounter = 0;
string Cache::CacheDir = ".cache/";

Cache::Cache()
{
  /* initialize member variables */
  m_vValue.clear();
  m_pSite = NULL;
  m_sFlightName = "";
}

Cache::~Cache()
{
  /* clear data vector */
  m_vValue.clear();

  /* free the site if allocated */
  if (m_pSite != NULL)
    {
      delete m_pSite;
      m_pSite = NULL;
    }
}

void Cache::init()
{
  int fd;

  /* attempt to make the cache directory */
  fd = mkdir((char*)CacheDir.c_str(), 0755);

  /* open the TOC file */
  fd = open((char*)(CacheDir+"/toc").c_str(), O_CREAT | O_RDWR);

  /* check for errors */
  if (fd == -1)
    {
      perror("Cache::init(): ");

      /* this is the first cache entry */
      CacheNameCounter = 0;
    }
  else 
    {
      /* make sure users can read/write this file */
      fchmod(fd, S_IRUSR | S_IWUSR);

      /* read the number of cache entries */
      if (read(fd, &CacheNameCounter, sizeof(unsigned long)) == 0)
	{
	  /* if the read fails, set this as the first entry */
	  CacheNameCounter = 0;

	  /* write the entry to the beginning of the file */
	  write(fd, &CacheNameCounter, sizeof(unsigned long));
	}

      /* close the TOC */
      close(fd);
    }
}

void Cache::updatetoc()
{
  int fd;

  /* open the TOC file */
  fd = open((char*)(CacheDir+"/toc").c_str(), O_RDWR | S_IRUSR | S_IWUSR);

  /* check for errors */
  if (fd == -1)
    perror("Cache::updatetoc(): ");
  else 
    {
      /* write the number of entries to the file */
      write(fd, &CacheNameCounter, sizeof(unsigned long));

      /* close the file */
      close(fd);
    }
}

bool Cache::LoadFromFile(char *fname)
{
  int fd;
  struct header hdr;
  unsigned long i, imax;
  char *buf1, *buf2;
  unsigned char b;
  char j, bmax;

  /* open cache file */
  fd = open(fname, O_RDONLY);

  /* check for errors */
  if (fd == -1)
    return false;

  /* read the header */
  read(fd, &hdr, sizeof(struct header));

  /* check the magic number for a valid file */
  if (hdr.magic != 0xdeadbeef)
    {
      /* close the file */
      close(fd);

      /* return failed */
      return false;
    }

  /* allocate buffer for header name */
  buf1 = new char[hdr.namelen+1];

  /* clear buffer */
  memset(buf1, 0, hdr.namelen+1);

  /* read name from file into buffer */
  read(fd, buf1, hdr.namelen);

  /* allocate buffer for flight name */
  buf2 = new char[hdr.flightlen+1];

  /* clear buffer */
  memset(buf2, 0, hdr.flightlen+1);

  /* read name from file into buffer */
  read(fd, buf2, hdr.flightlen);

  /* create new Site from the site parameters */
  m_pSite = new Site(buf1, 
		     hdr.site.lat, 
		     hdr.site.lon, 
		     hdr.site.alt, 
		     hdr.site.ant);

  /* set flight name */
  m_sFlightName = string(buf2);

  /* calculate number of bits per byte */
  bmax = 8*sizeof(unsigned char);

  /* calculate number of bytes to scan through */
  imax = hdr.points/8;

  /* sanity check for partially full bytes */
  if (hdr.points % 8 != 0)
    imax++;

  /* clear the data vector */
  m_vValue.clear();

  /* resize the data vector */
  Count(hdr.points);

  /* loop through data bytes */
  for (i = 0; i < imax; i++)
    {
      /* get next byte */
      read(fd, &b, sizeof(unsigned char));

      /* set/clear each bitvalue in the data vector */
      for (j = 0; j < bmax; j++)
	m_vValue[(i*bmax)+j] = ((b & (1<<j)) != 0);
    }

  /* close cache file */
  close(fd);

  /* cleanup buffers */
  delete [] buf1;
  delete [] buf2;

  /* return success */
  return true;
}

bool Cache::SaveToFile(char *fname)
{
  struct header hdr;
  int fd;
  unsigned long i, imax;
  unsigned char b;
  char j, bmax;
  string pstr;

  /* return if the base site is empty */
  if (m_pSite == NULL)
    return false;

  /* initialize cache directory string */
  pstr = CacheDir;
  
  /* set magic value */
  hdr.magic = 0xdeadbeef;

  /* set number of points */
  hdr.points = m_vValue.size();

  /* set site position */
  hdr.site.lat = m_pSite->Latitude();
  hdr.site.lon = m_pSite->Longitude();
  hdr.site.alt = m_pSite->Altitude();
  hdr.site.ant = m_pSite->Antenna();

  /* set site name length */
  hdr.namelen = m_pSite->Name().length();

  /* set flight name length */
  hdr.flightlen = m_sFlightName.length();

  /* figure out new filename if none provided */
  if (fname == NULL)
    {
      char buf[64];
      sprintf(buf, "/%06lu-cache.lch", CacheNameCounter++);
      pstr += string(buf);
    }
  else pstr += string("/")+string(fname);

  /* open cache file */
  fd = open(pstr.c_str(), O_CREAT | O_WRONLY | S_IREAD | S_IWRITE);

  /* check for errors */
  if (fd == -1)
    {
      /* display errors */
      cout<<"pstr = "<<pstr<<endl;
      perror("savetofile::open(): ");
      return false;
    }

  /* make sure users can read and write to the file */
  fchmod(fd, S_IRUSR | S_IWUSR);

  /* write header to file */
  write(fd, &hdr, sizeof(struct header));

  /* write site name to file */
  write(fd, m_pSite->Name().c_str(), hdr.namelen);

  /* write flight name to file */
  write(fd, m_sFlightName.c_str(), hdr.flightlen);

  /* calculate number of bits per byte */  
  bmax = 8*sizeof(unsigned char);

  /* calculate number of bytes for all points */
  imax = hdr.points/8;

  /* sanity check for partially full bytes */
  if (hdr.points % 8 != 0)
    imax++;

  /* iterate through the bytes */
  for (i = 0; i < imax; i++)
    {
      /* reset data byte */
      b = 0;

      /* set/clear each bit */
      for (j = 0; j < bmax; j++)
	{
	  if (m_vValue[(i*bmax)+j] == true)
	    b |= (1 << j);
	}

      /* write byte to file */
      write(fd, &b, sizeof(unsigned char));
    }

  /* close the file */
  close(fd);

  /* update the TOC file */
  updatetoc();

  /* return success */
  return true;
}

bool Cache::Value(unsigned long step)
{
  if (0 <= step && step < m_vValue.size())
    return m_vValue[step];
  else return false;
}

void Cache::Value(unsigned long step, bool val)
{
  if (0 <= step && step < m_vValue.size())
    m_vValue[step] = val;
}

void Cache::POV(Site &pov)
{
  if (m_pSite != NULL)
    delete m_pSite;
  m_pSite = new Site(pov);
}

int Cache::FindFiles(vector<string> &list)
{
  DIR *dp;
  struct dirent *de;
  string fname, basepath;
  bool found;

  /* open directory */
  dp = opendir(CacheDir.c_str());

  /* check for errors */
  if (dp == NULL)
    {
      perror("FindFile(): ");
      return 0;
    }

  /* clear file list */
  list.clear();

  /* reset local variables */
  fname = "";
  basepath = CacheDir+"/";
  found = false;

  /* iterate through files */
  while (!found && (de = readdir(dp)) != NULL)
    {
      /* make sure the file is a block file */
      if (de->d_type == DT_FIFO ||
	  de->d_type == DT_CHR ||
	  de->d_type == DT_DIR ||
	  de->d_type == DT_BLK ||
	  de->d_type == DT_SOCK ||
	  de->d_type == DT_WHT)
	continue;

      /* make sure the file is not the TOC file */
      if (!strncmp(de->d_name, "toc", 3) && strlen(de->d_name) == 3)
	continue;

      /* make sure the file suffix is .lch */
      if (strncmp(de->d_name+(strlen(de->d_name)-4), ".lch", 4) != 0)
	continue;

      /* add file name to list */
      list.push_back(basepath+de->d_name);
      /*printf("\tCache::FindFile(): %d %s\n", de->d_type, de->d_name);*/
    }

  /* close the directory */
  closedir(dp);

  /* return number of files found */
  return (int)list.size();
}

bool Cache::Find(Site *pov, char *flight_name, string &cache_name)
{
  vector<string> list;
  int i, fd;
  struct header hdr;
  char *buf1, *buf2;
  Site *s;
  string flightname_cached;
  string flightname_current;

  /* reset local variables */
  cache_name = "";
  flightname_current = string(flight_name);

  /* find files */
  FindFiles(list);
  
  /*cout<<"Cache::Find() returned "<<list.size()<<" files\n";
    cout<<"flightname_current: "<<flightname_current<<endl;*/

  /* iterate through the file list */
  for (i = 0; i < (int)list.size(); i++)
    {
      /* open the file */
      fd = open(list[i].c_str(), O_RDONLY);

      /* check for errors */
      if (fd == -1)
	{
	  perror("Cache::Find()::open: ");
	  continue;
	}

      /* read the header */
      read(fd, &hdr, sizeof(struct header));

      /* check if the file is good */
      if (hdr.magic != 0xdeadbeef)
	{
	  /* close the file */
	  close(fd);
	  continue;
	}

      /* allocate buffer for site name */
      buf1 = new char[hdr.namelen+1];

      /* clear buffer */
      memset(buf1, 0, hdr.namelen+1);

      /* read site name */
      read(fd, buf1, hdr.namelen);
      
      /* allocate buffer for flight name */
      buf2 = new char[hdr.flightlen+1];

      /* clear buffer */
      memset(buf2, 0, hdr.flightlen+1);
      
      /* read flight name */
      read(fd, buf2, hdr.flightlen);

      /* create new Site from site parameters */
      s = new Site(buf1, 
		   hdr.site.lat, 
		   hdr.site.lon, 
		   hdr.site.alt, 
		   hdr.site.ant);

      /* set cached flight name */
      flightname_cached = string(buf2);

      /* cleanup buffers */
      delete [] buf1;
      delete [] buf2;

      /*cout<<"flightname_cached:  "<<flightname_cached<<endl;
	cout<<"flightname_current: "<<flightname_current<<endl;
	cout<<"equals: "<<(flightname_cached == flightname_current)<<endl;
	cout<<endl;*/

      /* check if the sites are the same, as well as the flight names */
      if (*s == *pov && flightname_cached == flightname_current)
	{
	  /* free allocated site */
	  delete s;

	  /* set cache name to current file name */
	  cache_name = list[i];
	  
	  /*cout<<"matched site + filenames\n";*/

	  /* return success */
	  return true;
	}
    }

  /* return failed */
  return false;
}
