#include "dted.h"
#include "collection.h"
#include "coord.h"
#include "site.h"
#include "path.h"
#include "dtedgrid.h"
#include "geod.h"

Collection::Collection()
{
	Clear();
	m_pGrid = NULL;
}

Collection::Collection(char *fname)
{
	Clear();
	m_pGrid = NULL;

	if (!LoadSitesFromFile(fname))
		throw bad_exception();
}

Collection::~Collection()
{
	Clear();
}

bool Collection::GeneratePath(float inc, bool interpalt)
{
	int i;
	Path *p;

	if (m_vSites.size() < 2 || m_pGrid == NULL)
		return false;

	while (m_vPaths.size() != 0)
	{
		delete m_vPaths[0];
		m_vPaths.erase(m_vPaths.begin(), m_vPaths.begin()+1);
	}

	for (i = 0; i < (int)m_vSites.size()-1; i++)
	{
		p = new Path();
		p->CreatePath(m_vSites[i], m_vSites[i+1], (DTED*)NULL,
			m_pGrid->_dtedlookupfunc, false, 0.04f, interpalt);
		m_vPaths.push_back(p);
	}

	return true;
}


int Collection::AddSite(Site *v)
{
	/* check to see if site has already been stored */
	if (FindSite(v) != -1)
		return -1;

	m_vSites.push_back(v);
	return (int)m_vSites.size()-1;
}

void Collection::DelSite(int index)
{
	printf("DelSite(%d)\n", index);
}

int Collection::FindSite(Site *v)
{
	vector<Site*>::iterator p;

	for (p = m_vSites.begin(); p != m_vSites.end(); p++)
		if (*(*p) == *v)
			return (int)(p-m_vSites.begin());

	return -1;
}

void Collection::Clear()
{
	while (m_vSites.size() != 0)
	{
		delete m_vSites[0];
		m_vSites.erase(m_vSites.begin(), m_vSites.begin()+1);
	}

	while (m_vPaths.size() != 0)
	{
		delete m_vPaths[0];
		m_vPaths.erase(m_vPaths.begin(), m_vPaths.begin()+1);
	}
}

Site *Collection::Sites(int index)
{
	if (0 <= index && index < (int)m_vSites.size())
		return m_vSites[index];
	else return NULL;
}

Path *Collection::Paths(int index)
{
	if (0 <= index && index < (int)m_vPaths.size())
		return m_vPaths[index];
	else return NULL;
}

Site *Collection::GlobalSite(unsigned int index)
{
	unsigned int acc = 0;
	unsigned int pindex = 0;

	if (index >= TotalCoordCount())
		return NULL;

	while (acc < index && pindex < m_vPaths.size())
	{
		Path *p = m_vPaths[pindex];

		if (acc+p->Length() > index)
			return p->LocationSite(index-acc);

		acc += p->Length();
		pindex++;
	}
	return NULL;
}

long Collection::MaxPathAltitude()
{
	int i, j;
	long maxalt = -65535;
	Path *p;
	Coord *c;

	for (i = 0; i < (int)m_vPaths.size(); i++)
	{
		p = m_vPaths[i];
		for (j = 0; j < (int)p->Length(); j++)
		{
			c = p->Location(j);
			if ((long)c->Altitude > maxalt)
				maxalt = (long)c->Altitude;
		}
	}

	return maxalt;
}

long Collection::MinPathAltitude()
{
	int i, j;
	long minalt = 65535;
	Path *p;
	Coord *c;

	for (i = 0; i < (int)m_vPaths.size(); i++)
	{
		p = m_vPaths[i];
		for (j = 0; j < (int)p->Length(); j++)
		{
			c = p->Location(j);
			if ((long)c->Altitude < minalt)
				minalt = (long)c->Altitude;
		}
	}
	return minalt;
}

long Collection::MaxSiteAltitude()
{
	int i;
	long maxalt = -65535;
	Site *s;

	for (i = 0; i < (int)m_vSites.size(); i++)
	{
		s = m_vSites[i];
		if ((long)s->Altitude() > maxalt)
			maxalt = (long)s->Altitude();
	}

	return maxalt;
}

long Collection::MinSiteAltitude()
{
	int i;
	long minalt = 65535;
	Site *s;

	for (i = 0; i < (int)m_vSites.size(); i++)
	{
		s = m_vSites[i];
		if ((long)s->Altitude() < minalt)
			minalt = (long)s->Altitude();
	}

	return minalt;
}

bool Collection::LoadSitesFromFile(char *fname)
{
	ifstream ip(fname);
	string line;
	int site_index;
	double site_lat, site_lon, site_alt;
	string site_name;
	Site *c = NULL;
	string path = ExtractFilePath(fname);

	Clear();

	while (getline(ip, line, '\n'))
	{
		if (line[0] != '!')
			continue;

		if (line.find("!index:") == 0)
		{
			/* index */
			size_t length = strlen("!index: ");
			line = line.substr(length, line.length()-length);
			site_index = atoi(line.c_str());
			/*cout<<"index: "<<site_index<<endl;*/

			c = new Site();
		}
		else if (line.find("!name:") == 0)
		{
			/* index */
			size_t length = strlen("!name: ");
			line = line.substr(length, line.length()-length);
			site_name = line;
			/*cout<<"name: "<<site_name<<endl;*/

			c->Name(site_name);
		}
		else if (line.find("!latitude:") == 0)
		{
			/* latitude */
			size_t length = strlen("!latitude: ");
			line = line.substr(length, line.length()-length);
			site_lat = geod::str2pos(line);
			/*cout<<"latitude: "<<site_lat<<endl;*/

			c->Latitude(site_lat);
		}
		else if (line.find("!longitude:") == 0)
		{
			/* longitude */
			size_t length = strlen("!longitude: ");
			line = line.substr(length, line.length()-length);
			site_lon = geod::str2pos(line);
			/*cout<<"longitude: "<<site_lon<<endl;*/

			c->Longitude(site_lon);
		}
		else if (line.find("!altitude:") == 0)
		{
			/* altitude */
			size_t length = strlen("!altitude: ");
			line = line.substr(length, line.length()-length);
			site_alt = atof(line.c_str());
			/*cout<<"altitude: "<<site_alt<<endl;*/

			c->Altitude(site_alt);
			/*c->Print();*/
			m_vSites.push_back(c);

			/*cout<<"load ptr = "<<c<<endl;*/
			c = NULL;
		}
		else if (line.find("!file:") == 0)
		{
			/* site file */
			size_t length = strlen("!file: ");
			line = line.substr(length, line.length()-length);
			line = path+line;

			c = new Site((char*)line.c_str());
			m_vSites.push_back(c);
			c = NULL;
		}
		else if (line.find("!ignore:") == 0)
		{
			/* use directive */
			size_t length = strlen("!ignore: ");
			line = line.substr(length, line.length()-length);

			int ignore_index = atoi(line.c_str());
			cout<<"IGNORE directive: "<<ignore_index<<endl;

			if (ignore_index < (int)m_vSites.size())
				m_vSites[ignore_index]->Active(false);
		}
	}
	ip.close();

	return true;
}

bool Collection::SaveSitesToFile(char *fname)
{
	int i;
	ofstream op(fname);

	for (i = 0; i < (int)m_vSites.size(); i++)
	{
		Site *c;

		c = m_vSites[i];

		op<<"!index: "<<i<<endl;
		op<<"!name: "<<c->Name()<<endl;
		op<<"!latitude: "<<geod::pos2str(c->Latitude(), 0)<<endl;
		op<<"!longitude: "<<geod::pos2str(c->Longitude(), 1)<<endl;
		op<<"!altitude: "<<c->Altitude()<<endl;
		op<<endl;

		/*cout<<"save ptr["<<i<<"] = "<<m_vSites[i]<<endl;*/
	}
	op.close();

	return true;
}

#ifdef _INCLUDE_LOAD_SAVE_PATHS
bool Collection::LoadPathsFromFile(char *fname)
{
	printf("LoadPathsFromFile(%s)\n", fname);
	return true;
}

bool Collection::SavePathsToFile(char *fname)
{
	printf("SavePathsFromFile(%s)\n", fname);
	return true;
}

bool Collection::SavePathsForGnuplot(string base_fname, Site *ref)
{
	string path_fname = base_fname+".path";
	string ctrl_fname = base_fname+".ctrl";
	string waypt_fname = base_fname+".waypt";
	string sitealt_fname = base_fname+".level";

	FILE *fpp = fopen((char*)path_fname.c_str(), "w+");
	FILE *fpc = fopen((char*)ctrl_fname.c_str(), "w+");
	FILE *fpw = fopen((char*)waypt_fname.c_str(), "w+");
	FILE *fps = fopen((char*)sitealt_fname.c_str(), "w+");
	int i, j;
	unsigned long cnt;
	long maxpathalt, minpathalt;
	long maxsitealt, minsitealt;
	long maxalt, minalt;
	Path *p;
	Site *s;
	Coord *c;

	if (fpp == NULL || fpc == NULL || fpw == NULL || fps == NULL)
	{
		fclose(fpp);
		fclose(fpc);
		fclose(fpw);
		fclose(fps);
		return false;
	}

	/* find the maximum path altitude */
	maxpathalt = MaxPathAltitude();

	/* find the minimum path altitude */
	minpathalt = MinPathAltitude();

	/* find the maximum site altitude */
	maxsitealt = MaxSiteAltitude();

	/* find the minimum site altitude */
	minsitealt = MinSiteAltitude();

	if (maxsitealt > maxpathalt)
		maxalt = maxsitealt;
	else maxalt = maxpathalt;

	if (minsitealt < minpathalt)
		minalt = minsitealt;
	else minalt = minpathalt;

	fprintf(fpc, "!set nokey\n");
	fprintf(fpc, "set yrange [%ld:%ld]\n", minalt-50, maxalt+50);
	fprintf(fpc, "set terminal postscript color\n");
	fprintf(fpc, "set output \"%s.ps\"\n", base_fname.c_str());
	fprintf(fpc, "set xlabel \"Path Distance [100m]\"\n");
	fprintf(fpc, "set ylabel \"Altitude [m]\"\n");

	cnt = 0;
	/* for each path object... */
	for (i = 0; i < (int)m_vPaths.size(); i++)
	{
		p = m_vPaths[i];
		s = m_vSites[i];

		if (ref != NULL && i == 0)
			fprintf(fpw, "%lu\t%ld\t%ld\t%ld\n", cnt, minalt-50, (long)s->Altitude(), (long)ref->Altitude());
		else fprintf(fpw, "%lu\t%ld\t%ld\n", cnt, minalt-50, (long)s->Altitude());
		fprintf(fpw, "%lu\t%lu\n", cnt, maxalt+50);
		fprintf(fpw, "%lu\t%ld\n", cnt, minalt-50);

		fprintf(fpc, "set label %d \"%s\" at %lu,%ld center font \"Helvetica,8\"\n", i+1, s->Name().c_str(), cnt, maxalt+97);
		for (j = 0; j < (int)p->Length(); j++)
		{
			c = p->Location(j);
			fprintf(fpp, "%lu\t%f\n", cnt++, c->Altitude);
		}
	}

	/* final site (one more than the number of path objects) */
	s = m_vSites[m_vSites.size()-1];
	fprintf(fpw, "%lu\t%ld\n", cnt, minalt-50);
	fprintf(fpw, "%lu\t%lu\n", cnt, maxalt+50);
	if (ref != NULL)
		fprintf(fpw, "%lu\t%ld\t%ld\t%ld\n", cnt, minalt-50, (long)s->Altitude(), (long)ref->Altitude());
	else fprintf(fpw, "%lu\t%ld\t%ld\n", cnt, minalt-50, (long)s->Altitude());

	fprintf(fpc, "set label %d \"%s\" at %lu,%ld center font \"Helvetica,8\"\n", i+1, s->Name().c_str(), cnt, maxalt+97);

	fprintf(fpc, "set label \"GPS/INS 471200D\" at screen 0.0125,0.0275 font \"Helvetica,8\"\n");
	fprintf(fpc, "set label \"James Rogers, 939-0103\" at screen 0.0125,0.0125 font \"Helvetica,8\"\n");

	if (ref != NULL)
		fprintf(fpc, "plot \"%s\" title \"Terrain\" with lines, \"%s\" using 1:2 title \"Waypoints\" with lines, \"%s\" using 1:3 title \"Path\" with lines, \"%s\" using 1:4 title \"%s\" with lines\n",
		path_fname.c_str(), waypt_fname.c_str(), waypt_fname.c_str(), waypt_fname.c_str(), ref->Name().c_str());
	else fprintf(fpc, "plot \"%s\" title \"Terrain\" with lines, \"%s\" using 1:2 title \"Waypoints\" with lines, \"%s\" using 1:3 title \"Path\" with lines\n",
		path_fname.c_str(), waypt_fname.c_str(), waypt_fname.c_str());

	fclose(fpp);
	fclose(fpc);
	fclose(fpw);
	fclose(fps);

	return true;
}
#endif

void Collection::CoordBounds(Site *extra,
							 double *lonmin, double *latmin,
							 double *lonmax, double *latmax,
							 bool verbose)
{
	int i;

	*lonmin = 360;
	*lonmax = -360;
	*latmin = 90;
	*latmax = -90;

	for (i = 0; i < (int)m_vSites.size(); i++)
	{
		Site *s;

		s = m_vSites[i];

		if (s->Latitude() < *latmin)
			*latmin = s->Latitude();
		if (s->Latitude() > *latmax)
			*latmax = s->Latitude();

		if (s->Longitude() < *lonmin)
			*lonmin = s->Longitude();
		if (s->Longitude() > *lonmax)
			*lonmax = s->Longitude();
	}

	if (extra != NULL)
	{
		if (extra->Latitude() < *latmin)
			*latmin = extra->Latitude();
		if (extra->Latitude() > *latmax)
			*latmax = extra->Latitude();

		if (extra->Longitude() < *lonmin)
			*lonmin = extra->Longitude();
		if (extra->Longitude() > *lonmax)
			*lonmax = extra->Longitude();
	}
	if (verbose)
		printf("cb::bounds: min=(%d, %d), max=(%d, %d)\n",
		(int)*lonmin, (int)*latmin,
		(int)*lonmax, (int)*latmax);

	*lonmin = floor(*lonmin);
	*lonmax = ceil(*lonmax);
	*latmin = floor(*latmin);
	*latmax = ceil(*latmax);
}

void Collection::CoordBounds(Collection *extra,
							 double *xmin, double *ymin,
							 double *xmax, double *ymax,
							 bool verbose)
{
	int i;

	*xmin = 360;
	*xmax = -360;
	*ymin = 90;
	*ymax = -90;

	CoordBounds((Site*)NULL, xmin, ymin, xmax, ymax, verbose);

	if (verbose)
		printf("CoordBounds::SC.S::bounds: min=(%f, %f), max=(%f, %f)\n",
		*xmin, *ymin, *xmax, *ymax);

	if (extra != NULL)
	{
		for (i = 0; i < (int)extra->SiteCount(); i++)
		{
			Site *s = extra->Sites(i);

			if (s->Latitude() < *ymin)
				*ymin = s->Latitude();
			if (s->Latitude() > *ymax)
				*ymax = s->Latitude();

			if (s->Longitude() < *xmin)
				*xmin = s->Longitude();
			if (s->Longitude() > *xmax)
				*xmax = s->Longitude();
		}
	}

	if (verbose)
		printf("CoordBounds::SC.C::bounds: min=(%f, %f), max=(%f, %f)\n",
		*xmin, *ymin, *xmax, *ymax);
	*xmin = floor(*xmin);
	*xmax = ceil(*xmax);
	*ymin = floor(*ymin);
	*ymax = ceil(*ymax);
	if (verbose)
		printf("CoordBounds::SC.C::bounds: min=(%f, %f), max=(%f, %f)\n",
		*xmin, *ymin, *xmax, *ymax);
}

#ifndef _NO_PLOTXY
void Collection::GraphicBounds(Site *extra,
							   int &xmin, int &ymin, 
							   int &xmax, int &ymax)
{
	int i;
	double dxmin, dymin, dxmax, dymax;

	if (m_pGrid == NULL)
		return;

	dxmin = 360;
	dxmax = -360;
	dymin = 90;
	dymax = -90;

	for (i = 0; i < (int)m_vSites.size(); i++)
	{
		Site *s = m_vSites[i];

		if (s->Latitude() < dymin)
			dymin = s->Latitude();
		if (s->Latitude() > dymax)
			dymax = s->Latitude();

		if (s->Longitude() < dxmin)
			dxmin = s->Longitude();
		if (s->Longitude() > dxmax)
			dxmax = s->Longitude();
	}

	if (extra != NULL)
	{
		if (extra->Latitude() < dymin)
			dymin = extra->Latitude();
		if (extra->Latitude() > dymax)
			dymax = extra->Latitude();

		if (extra->Longitude() < dxmin)
			dxmin = extra->Longitude();
		if (extra->Longitude() > dxmax)
			dxmax = extra->Longitude();
	}

	int xc1, yc1, xc2, yc2;

	/* Map lat/lon coordinates to cell indices in the grid */
	DTED *d1 = m_pGrid->MapCoordToCell(dymin, dxmin, &xc1, &yc1);
	DTED *d2 = m_pGrid->MapCoordToCell(dymax, dxmax, &xc2, &yc2);

	if (d1 != NULL && d2 != NULL)
	{
		/* Map lat/lon coordinates to post indices in the cell */
		d1->MapCoordToPost(dymin, dxmin, xmin, ymin);
		d2->MapCoordToPost(dymax, dxmax, xmax, ymax);

		/* Map post indices to graphic coordinates */
#if 1
		d1->MapPostToGraphic(xmin, ymin, -1, -1);
		d2->MapPostToGraphic(xmax, ymax, -1, -1);
#else
		d1->MapPostToGraphic(xmin, ymin, m_pGrid->PlotXMin(), m_pGrid->PlotYMin());
		d2->MapPostToGraphic(xmax, ymax, m_pGrid->PlotXMin(), m_pGrid->PlotYMin());
#endif
	}
}

void Collection::GraphicBounds(Collection *extra,
							   int &xmin, int &ymin, 
							   int &xmax, int &ymax)
{
	int i;
	double dxmin, dymin, dxmax, dymax;

	if (m_pGrid == NULL)
		return;

	dxmin = 360;
	dxmax = -360;
	dymin = 90;
	dymax = -90;

	for (i = 0; i < (int)m_vSites.size(); i++)
	{
		Site *s = m_vSites[i];

		if (s->Latitude() < dymin)
			dymin = s->Latitude();
		if (s->Latitude() > dymax)
			dymax = s->Latitude();

		if (s->Longitude() < dxmin)
			dxmin = s->Longitude();
		if (s->Longitude() > dxmax)
			dxmax = s->Longitude();
	}

	if (extra != NULL)
	{
		for (i = 0; i < (int)extra->SiteCount(); i++)
		{
			Site *s = extra->Sites(i);

			if (s->Latitude() < dymin)
				dymin = s->Latitude();
			if (s->Latitude() > dymax)
				dymax = s->Latitude();

			if (s->Longitude() < dxmin)
				dxmin = s->Longitude();
			if (s->Longitude() > dxmax)
				dxmax = s->Longitude();
		}
	}

	int xc1, yc1, xc2, yc2;

	/* Map lat/lon coordinates to cell indices in the grid */
	DTED *d1 = m_pGrid->MapCoordToCell(dymin, dxmin, &xc1, &yc1);
	DTED *d2 = m_pGrid->MapCoordToCell(dymax, dxmax, &xc2, &yc2);

	if (d1 != NULL && d2 != NULL)
	{
		/* Map lat/lon coordinates to post indices in the cell */
		d1->MapCoordToPost(dymin, dxmin, xmin, ymin);
		d2->MapCoordToPost(dymax, dxmax, xmax, ymax);

		/* Map post indices to graphic coordinates */
#if 1
		d1->MapPostToGraphic(xmin, ymin, -1, -1);
		d2->MapPostToGraphic(xmax, ymax, -1, -1);
#else
		d1->MapPostToGraphic(xmin, ymin, m_pGrid->PlotXMin(), m_pGrid->PlotYMin());
		d2->MapPostToGraphic(xmax, ymax, m_pGrid->PlotXMin(), m_pGrid->PlotYMin());
#endif
	}
}
#endif

unsigned int Collection::TotalCoordCount()
{
	unsigned int acc = 0;
	unsigned int i;

	for (i = 0; i < m_vPaths.size(); i++)
	{
		Path *p = m_vPaths[i];
		acc += p->Length();
	}

	return acc;
}

string Collection::ExtractFilePath(char *fname)
{
	char *p;
	string str;

#ifdef WIN32
	p = strrchr(fname, '\\');
	if (p)
	{
		str = fname;
		str = str.substr(0, p-fname);
	}
	str += "\\";
#else
	p = strrchr(fname, '/');
	if (p)
	{
		str = fname;
		str = str.substr(0, p-fname);
	}
	str += "/";
#endif
		
	return str;
}
