#include <cstring>
#include <plotter.h>
#include "contour.h"

Contour::Contour(string fname) :
  m_fScaleWidth(1.0),
  m_fScaleHeight(1.0),
  m_iWidth(0),
  m_iHeight(0),
  m_iContourCount(0),
  m_pfLevels(NULL),
  m_pfXCoords(NULL),
  m_pfYCoords(NULL),
  m_ppfZData(NULL),
  m_sFileName(fname),
  m_pCallbackPtr(NULL),
  m_pCallbackParam(NULL)
{
}

Contour::~Contour()
{
  /* clean up memory */
  _Deallocate(true);
}

int Contour::Calculate(string &fname, DrawCallback cb, void *param)
{
  int r;
  const unsigned long magic = 0xdeadbeef;
  
  m_pCallbackPtr = cb;
  m_pCallbackParam = param;
  
  if (!m_pfLevels)
    return -1;
  if (!m_pfXCoords)
    return -2;
  if (!m_pfYCoords)
    return -3;
  if (!m_ppfZData)
    return -4;

  if (fname != "")
    m_sFileName = fname;
  
  if (m_sFileName == "")
    m_pLogFile = NULL;
  else m_pLogFile = fopen(m_sFileName.c_str(), "w+");
  
  fwrite(&magic, sizeof(unsigned long), 1, m_pLogFile);
  fwrite(&m_iWidth, sizeof(int), 1, m_pLogFile);
  fwrite(&m_iHeight, sizeof(int), 1, m_pLogFile);  

  r = _calculate();
  
  if (m_pLogFile)
    fclose(m_pLogFile);

  return r;
}

bool Contour::Create(int width, int height, int xmin, int ymin)
{
  /* set minimums for X and Y */
  m_iXMin = xmin;
  m_iYMin = ymin;

  /* set width/height and allocate */
  _Allocate(width, height);

  /* return OK if all pointers are non-NULL */
  return (m_pfXCoords != NULL && m_pfYCoords != NULL &&
	  m_ppfZData != NULL);
}

void Contour::_Allocate(int width, int height)
{
  int i;

  /* deallocate anything already allocated */
  _Deallocate(true);

  /* set width */
  m_iWidth = width;

  /* set height */
  m_iHeight = height;

  /* allocate new buffers for width/height */
  m_pfXCoords = new float[m_iWidth];
  m_pfYCoords = new float[m_iHeight];
  
  /* allocate new buffer for m_iHeight float* values */
  m_ppfZData = new float*[m_iHeight];

  for (i = 0; i < m_iHeight; i++)
    {
      /* allocate each float* value */
      m_ppfZData[i] = new float[m_iWidth];

      /* set all locations to 0.0 */
      memset(m_ppfZData[i], 0, m_iWidth*sizeof(float));
    }
}

void Contour::_Deallocate(bool all)
{
  int i;

  /* deallocate each double* value */
  for (i = 0; i < m_iHeight; i++)
    if (m_ppfZData[i])
      delete m_ppfZData[i];

  /* delete the double* array */
  if (m_ppfZData)
    {
      delete [] m_ppfZData;
      m_ppfZData = NULL;
    }

  /* check to see if EVERYTHING should be cleaned up */
  if (all)
    {
      /* free x-coordinate array */
      if (m_pfXCoords)
	{
	  delete [] m_pfXCoords;
	  m_pfXCoords = NULL;
	}

      /* free y-coordinate array */
      if (m_pfYCoords)
	{
	  delete [] m_pfYCoords;
	  m_pfYCoords = NULL;
	}

      /* free leve array */
      if (m_pfLevels)
	{
	  delete [] m_pfLevels;
	  m_pfLevels = NULL;
	}
    }
}

void Contour::Clear()
{
  int i;

  /* zero all height data */
  for (i = 0; i < m_iHeight; i++)
    memset(m_ppfZData[i], 0, m_iWidth*sizeof(double));
}

void Contour::ZData(int r, int c, float v)
{
  /* set height at a certain location */
  if ((0 <= r && r < m_iHeight) &&
      (0 <= c && c < m_iWidth))
    m_ppfZData[r][c] = v;
}

void Contour::_setcount(int v)
{
  m_iContourCount = v;
  
  if (m_pfLevels)
    delete [] m_pfLevels;
  
  m_pfLevels = new float[v];
}

int Contour::_calculate()
{
  //=============================================================================
  //
  //     CONREC is a contouring subroutine for rectangularily spaced data.
  //
  //     It emits calls to a line drawing subroutine supplied by the user
  //     which draws a contour map corresponding to real*4data on a randomly
  //     spaced rectangular grid. The coordinates emitted are in the same
  //     units given in the x() and y() arrays.
  //
  //     Any number of contour levels may be specified but they must be
  //     in order of increasing value.
  //
  //     As this code is ported from FORTRAN-77, please be very careful of the
  //     various indices like ilb,iub,jlb and jub, remember that C/C++ indices
  //     starts from zero (0)
  //
  //=============================================================================
  // d               ! matrix of data to contour              m_ppfZData
  // ilb,iub,jlb,jub ! index bounds of data matrix            0, Height, 0, Width
  // x               ! data matrix column coordinates         m_pfXCoords
  // y               ! data matrix row coordinates            m_pfYCoords
  // nc              ! number of contour levels               m_iContourCount
  // z               ! contour levels in increasing order     m_pfLevels

  int m1, m2, m3, case_value;
  double dmin, dmax/*, x1, x2, y1, y2*/;
  CallbackData draw;
  register int i, j, k, m;
  double h[5];
  int sh[5];
  double xh[5], yh[5];
  int h_ub, w_ub;

  /* set width and height */
  h_ub = m_iHeight-1;
  w_ub = m_iWidth-1;

  /* set 'thog' member of the draw variable to 'this' */
  draw.thog = this;
  
  //===========================================================================
  // The indexing of im and jm should be noted as it has to start from zero
  // unlike the fortran counter part
  //===========================================================================
  int im[4] = {0, 1, 1, 0};
  int jm[4] = {0, 0, 1, 1};

  //===========================================================================
  // Note that castab is arranged differently from the FORTRAN code because
  // Fortran and C/C++ arrays are transposed of each other, in this case
  // it is more tricky as castab is in 3 dimension
  //===========================================================================
  int castab[3][3][3] = { { {0, 0, 8}, {0, 2, 5}, {7, 6, 9} },
			  { {0, 3, 4}, {1, 3, 1}, {4, 3, 0} },
			  { {9, 6, 7}, {5, 2, 0}, {8, 0, 0} } };

  for (j = (w_ub-1); j >= 0; j--) {
    for (i = 0; i <= h_ub-1; i++) {
      double temp1,temp2;
      
      temp1 = min(m_ppfZData[i][j], m_ppfZData[i][j+1]);
      temp2 = min(m_ppfZData[i+1][j], m_ppfZData[i+1][j+1]);
      dmin = min(temp1,temp2);
      temp1 = max(m_ppfZData[i][j], m_ppfZData[i][j+1]);
      temp2 = max(m_ppfZData[i+1][j], m_ppfZData[i+1][j+1]);
      dmax = max(temp1, temp2);

      if (dmax >= m_pfLevels[0] && dmin <= m_pfLevels[m_iContourCount-1]) {
	for (k = 0; k < m_iContourCount; k++) {
	  if (m_pfLevels[k] >= dmin && m_pfLevels[k] <= dmax) {
	    for (m = 4; m >= 0; m--) {
	      if (m>0)
		{
		  //=============================================================
		  // The indexing of im and jm should be noted as it has to
		  // start from zero
		  //=============================================================
		  h[m] = m_ppfZData[i+im[m-1]][j+jm[m-1]]-m_pfLevels[k];
		  xh[m] = m_pfXCoords[i+im[m-1]];
		  yh[m] = m_pfYCoords[j+jm[m-1]];
		}
	      else
		{
		  h[0] = 0.25*(h[1]+h[2]+h[3]+h[4]);
		  xh[0] = 0.5*(m_pfXCoords[i]+m_pfXCoords[i+1]);
		  yh[0] = 0.5*(m_pfYCoords[j]+m_pfYCoords[j+1]);
		}

	      if (h[m] > 0.0)
		{
		  sh[m] = 1;
		}
	      else if (h[m] < 0.0)
		{
		  sh[m] = -1;
		}
	      else sh[m] = 0;
	    }
	    //=================================================================
	    //
	    // Note: at this stage the relative heights of the corners and the
	    // centre are in the h array, and the corresponding coordinates are
	    // in the xh and yh arrays. The centre of the box is indexed by 0
	    // and the 4 corners by 1 to 4 as shown below.
	    // Each triangle is then indexed by the parameter m, and the 3
	    // vertices of each triangle are indexed by parameters m1,m2,and
	    // m3.
	    // It is assumed that the centre of the box is always vertex 2
	    // though this is important only when all 3 vertices lie exactly on
	    // the same contour level, in which case only the side of the box
	    // is drawn.
	    //
	    //
	    //      vertex 4 +-------------------+ vertex 3
	    //               | \               / |
	    //               |   \    m-3    /   |
	    //               |     \       /     |
	    //               |       \   /       |
	    //               |  m=2    X   m=2   |       the centre is vertex 0
	    //               |       /   \       |
	    //               |     /       \     |
	    //               |   /    m=1    \   |
	    //               | /               \ |
	    //      vertex 1 +-------------------+ vertex 2
	    //
	    //
	    //
	    //               Scan each triangle in the box
	    //
	    //=================================================================
	    for (m = 1; m <= 4; m++)
	      {
		m1 = m;
		m2 = 0;
		
		if (m != 4)
		  m3 = m+1;
	        else m3 = 1;

	        case_value = castab[sh[m1]+1][sh[m2]+1][sh[m3]+1];
	        if (case_value != 0)
		  {
		    switch (case_value)
		      {
			//===========================================================
			//     Case 1 - Line between vertices 1 and 2
			//===========================================================
		      case 1:
			draw.x1 = xh[m1];
			draw.y1 = yh[m1];
			draw.x2 = xh[m2];
			draw.y2 = yh[m2];
			break;
			//===========================================================
			//     Case 2 - Line between vertices 2 and 3
			//===========================================================
		      case 2:
			draw.x1 = xh[m2];
			draw.y1 = yh[m2];
			draw.x2 = xh[m3];
			draw.y2 = yh[m3];
			break;
			//===========================================================
			//     Case 3 - Line between vertices 3 and 1
			//===========================================================
		      case 3:
			draw.x1 = xh[m3];
			draw.y1 = yh[m3];
			draw.x2 = xh[m1];
			draw.y2 = yh[m1];
			break;
			//===========================================================
			//     Case 4 - Line between vertex 1 and side 2-3
			//===========================================================
		      case 4:
			draw.x1 = xh[m1];
			draw.y1 = yh[m1];
			draw.x2 = xsect(m2, m3);
			draw.y2 = ysect(m2, m3);
			break;
			//===========================================================
			//     Case 5 - Line between vertex 2 and side 3-1
			//===========================================================
		      case 5:
			draw.x1 = xh[m2];
			draw.y1 = yh[m2];
			draw.x2 = xsect(m3, m1);
			draw.y2 = ysect(m3, m1);
			break;
			//===========================================================
			//     Case 6 - Line between vertex 3 and side 1-2
			//===========================================================
		      case 6:
			draw.x1 = xh[m1];
			draw.y1 = yh[m1];
			draw.x2 = xsect(m1, m2);
			draw.y2 = ysect(m1, m2);
			break;
			//===========================================================
			//     Case 7 - Line between sides 1-2 and 2-3
			//===========================================================
		      case 7:
			draw.x1 = xsect(m1, m2);
			draw.y1 = ysect(m1, m2);
			draw.x2 = xsect(m2, m3);
			draw.y2 = ysect(m2, m3);
			break;
			//===========================================================
			//     Case 8 - Line between sides 2-3 and 3-1
			//===========================================================
		      case 8:
			draw.x1 = xsect(m2, m3);
			draw.y1 = ysect(m2, m3);
			draw.x2 = xsect(m3, m1);
			draw.y2 = ysect(m3, m1);
			break;
			//===========================================================
			//     Case 9 - Line between sides 3-1 and 1-2
			//===========================================================
		      case 9:
			draw.x1 = xsect(m3, m1);
			draw.y1 = ysect(m3, m1);
			draw.x2 = xsect(m1, m2);
			draw.y2 = ysect(m1, m2);
			break;
		      default:
			break;
		      }
		    //=============================================================
		    // Put your processing code here and comment out the printf
		    //=============================================================
		    /* set altitude */
		    draw.z = m_pfLevels[k];
		    
		    /* save data to file */
		    if (m_pLogFile)
		      _SaveSegment(&draw);

		    /* callback function with values */
		    m_pCallbackPtr(&draw, m_pCallbackParam);
		  }
	      }
	  }
	}
      }
    }
  }
  return 0;
}
//---------------------------------------------------------------------------
void Contour::_SaveSegment(CallbackData *data)
{
  struct _pos
  {
    float x1, y1;
    float x2, y2;
    float z;
  } p;
  
  p.x1 = data->y1;
  p.y1 = data->x1;
  p.x2 = data->y2;
  p.y2 = data->x2;
  p.z = data->z;
  
  fwrite(&p, sizeof(struct _pos), 1, m_pLogFile);
}
//---------------------------------------------------------------------------
