#ifndef WIN32
#include <dirent.h>
#else
#include <io.h>
#endif
#include "dted.h"
#include "coord.h"
#include "site.h"

#define SMOOTH_ALTITUDE false

int DTED::PointCount[3] = { 0, 1201, 2402 };
int DTED::GridSize[3] = { 0, 100, 30 };

string DTED::PlotOriginLat = "";
string DTED::PlotOriginLon = "";
FloatCallback DTED::LoadStatusCB = NULL;
FloatCallback DTED::GraphStatusCB = NULL;

DTED::DTED(char *fname)
{
	m_bBigEndian = !_endian();
	m_iLatOffset = 0;
	m_iLonOffset = 0;
	m_iMinAltitude = 65535;
	m_iMaxAltitude = -65535;
	m_bLoaded = loadfromfile(fname);

	m_iPlotXMin = -1;
	m_iPlotYMin = -1;
	m_iPlotXMax = -1;
	m_iPlotYMax = -1;
	m_iPlotZoomLevel = 1;
	m_bPlotFake = false;
}

DTED::DTED(string &fname)
{
	m_bBigEndian = !_endian();
	m_iLatOffset = 0;
	m_iLonOffset = 0;
	m_bLoaded = loadfromfile(fname.c_str());

	m_iPlotXMin = -1;
	m_iPlotYMin = -1;
	m_iPlotXMax = -1;
	m_iPlotYMax = -1;
	m_iPlotZoomLevel = 1;
	m_bPlotFake = false;
}

DTED::~DTED()
{
	while ((int)m_vData.size() != 0)
	{
		/* TODO: fix this (segfaults) */
		delete [] m_vData[0].data;
		m_vData.erase(m_vData.begin(), m_vData.begin()+1);
	}
}


bool DTED::readheader()
{
	size_t bcnt;
#ifdef _SHOW_HEADER
	int i, latval, lonval;
	char buf[256];
#endif

	bcnt = fread(&m_Header.uhl.raw, sizeof(m_Header.uhl.raw), 1, m_pFilePtr);
	bcnt += fread(&m_Header.dsi.raw, sizeof(m_Header.dsi.raw), 1, m_pFilePtr);
	bcnt += fread(&m_Header.acc, sizeof(m_Header.acc), 1, m_pFilePtr);

	if (bcnt != 3)
		return false;

	m_sLevel = m_Header.dsi.raw.series_des[4]-'0';

	/* UHL processed data */
	m_Header.uhl.proc.series = m_Header.dsi.raw.series_des[4]-'0';
	m_Header.uhl.proc.origin_lat = str2lat(m_Header.uhl.raw.origin_lat, 8*sizeof(char));
	m_Header.uhl.proc.origin_lon = str2lon(m_Header.uhl.raw.origin_lon, 8*sizeof(char));
	sscanf(m_Header.uhl.raw.interval_lat, "%4d", &m_Header.uhl.proc.interval_lat);
	sscanf(m_Header.uhl.raw.interval_lon, "%4d", &m_Header.uhl.proc.interval_lon);
	sscanf(m_Header.uhl.raw.bandcount_lat, "%4d", &m_Header.uhl.proc.bandcount_lat);
	sscanf(m_Header.uhl.raw.bandcount_lon, "%4d", &m_Header.uhl.proc.bandcount_lon);

	/* DSI processed data */
	m_Header.dsi.proc.series_des = string(m_Header.dsi.raw.series_des, 5);
	m_Header.dsi.proc.origin_lat = str2lat(m_Header.dsi.raw.origin_lat, 9*sizeof(char));
	m_Header.dsi.proc.origin_lon = str2lon(m_Header.dsi.raw.origin_lon, 10*sizeof(char));
	m_Header.dsi.proc.sw_lat = (float)str2lat(m_Header.dsi.raw.sw_lat, 7*sizeof(char));
	m_Header.dsi.proc.sw_lon = (float)str2lon(m_Header.dsi.raw.sw_lon, 8*sizeof(char));
	m_Header.dsi.proc.nw_lat = (float)str2lat(m_Header.dsi.raw.nw_lat, 7*sizeof(char));
	m_Header.dsi.proc.nw_lon = (float)str2lon(m_Header.dsi.raw.nw_lon, 8*sizeof(char));
	m_Header.dsi.proc.ne_lat = (float)str2lat(m_Header.dsi.raw.ne_lat, 7*sizeof(char));
	m_Header.dsi.proc.ne_lon = (float)str2lon(m_Header.dsi.raw.ne_lon, 8*sizeof(char));
	m_Header.dsi.proc.se_lat = (float)str2lat(m_Header.dsi.raw.se_lat, 7*sizeof(char));
	m_Header.dsi.proc.se_lon = (float)str2lon(m_Header.dsi.raw.se_lon, 8*sizeof(char));
	sscanf(m_Header.dsi.raw.interval_lat, "%4d", &m_Header.dsi.proc.interval_lat);
	sscanf(m_Header.dsi.raw.interval_lon, "%4d", &m_Header.dsi.proc.interval_lon);
	sscanf(m_Header.dsi.raw.bandcount_lat, "%4d", &m_Header.dsi.proc.bandcount_lat);
	sscanf(m_Header.dsi.raw.bandcount_lon, "%4d", &m_Header.dsi.proc.bandcount_lon);
	sscanf(m_Header.dsi.raw.partialcell_indic, "%2hd", &m_Header.dsi.proc.partial);


#ifdef _SHOW_HEADER
	cout<<"DTED Info\n---------\n";
	cout<<"Origin latitude:      "<<m_Header.uhl.proc.origin_lat<<endl;
	cout<<"Origin longitude:     "<<m_Header.uhl.proc.origin_lon<<endl;
	cout<<"Longitude band count: "<<m_Header.uhl.proc.bandcount_lat<<endl;
	cout<<"Latitude band count:  "<<m_Header.uhl.proc.bandcount_lat<<endl;
	cout<<"Series:               "<<m_Header.uhl.proc.series<<endl;
	cout<<"Interval latitude:    "<<m_Header.dsi.proc.interval_lat<<endl;
	cout<<"Interval longitude:   "<<m_Header.dsi.proc.interval_lon<<endl;
	cout<<endl;
#endif

	return true;
}

/* Reads a DTED longitudinal band, storing the elevation data in *dat, and swapping
the data byte values if 'swapbytes' is true. */
int DTED::readband(dat_struct *dat)
{
	int i;
	size_t r;

	r = fread(&dat->hdr, sizeof(dat->hdr), 1, m_pFilePtr);
	if (r == 0 || dat->count == 0)
		return -1;

	dat->data = new short[dat->count];
	r = fread(dat->data, sizeof(short), dat->count, m_pFilePtr);

	dat->valid = (dat->hdr.sentinel == DAT_SENTINEL);
	fread(&dat->checksum, 4*sizeof(unsigned char), 1, m_pFilePtr);

	if (dat->valid)
	{
		if (m_bBigEndian == false)
		{
			for (i = 0; i < dat->count; i++)
			{
				unsigned short d;

				d = dat->data[i];
				dat->data[i] = ((d & 0x00ff) << 8) |
							   ((d & 0xff00) >> 8);
				//dat->data[i] = ntohs(dat->data[i]);
			}
		}

		/* values are signed magnitude, NOT two's complement */
		for (i = 0; i < dat->count; i++)
		{
			if ((dat->data[i] & (1<<15)) != 0)
			{
				dat->data[i] &= ~(1<<15);
				dat->data[i] *= -1;
			}

			/* determine min/max altitudes */
			if (dat->data[i] < m_iMinAltitude)
				m_iMinAltitude = dat->data[i];
			if (dat->data[i] > m_iMaxAltitude)
				m_iMaxAltitude = dat->data[i];
		}
	}
	else
	{
		printf("sentinel not found: %02x %02x %02x %02x %02x %02x %02x %02x\n",
			dat->hdr.sentinel,
			dat->hdr.block_count[0], dat->hdr.block_count[1], dat->hdr.block_count[2],
			dat->hdr.lon_count[0], dat->hdr.lon_count[1],
			dat->hdr.lat_count[0], dat->hdr.lat_count[1]);
	}

	return (int)r;
}

bool DTED::loadfromfile(const char *fname)
{
	bool ret;
	int i, count;
	dat_struct dat;

	m_vData.clear();
	m_iMinAltitude = 65535;
	m_iMaxAltitude = -65535;

#ifdef WIN32
	m_pFilePtr = fopen(fname, "rb");
#else
	m_pFilePtr = fopen(fname, "r");
#endif
	if (!m_pFilePtr)
		return false;

	ret = readheader();
	if (ret == false)
	{
		fclose(m_pFilePtr);
		return false;
	}

	ret = true;
	count = (m_sLevel == 1)?LVL1_COUNT:(m_sLevel == 2)?LVL2_COUNT:0;
	for (i = 0; i < count; i++)
	{
		dat.level = m_sLevel;
		dat.count = count;
		dat.valid = false;

		readband(&dat);
		if (dat.valid > 0)
			m_vData.push_back(dat);
		else
		{
			//cout<<"invalid: chunk "<<i<<", returning\n";
			ret = false;
			break;
		}

		if (LoadStatusCB)
			LoadStatusCB(this, ((float)i/(float)count)*100.0f);
	}

	if (LoadStatusCB)
		LoadStatusCB(this, 100.0);

	fclose(m_pFilePtr);

	return ret;
}

bool DTED::_endian()
{
	union probe
	{
		unsigned int num;
		unsigned char bytes[sizeof(unsigned int)];
	} p= { 1U };
	return (p.bytes[0] == 1U);
}

short DTED::AltitudeByPost(int x, int y)
{
	if (0 <= x && x < (int)m_vData.size())
	{
		if (0 <= y && y < m_vData[0].count)
			return m_vData[x].data[y];
		else return -5000;
	}
	else return -400;
}

short DTED::AltitudeByGeodetic(double lat, double lon, bool interpolate, int radius)
{
	int cell_x, cell_y;

	MapCoordToPost(lat, lon, cell_x, cell_y/*, interpolate*/);

	if (radius == 0)
	{
		/* just this point */

		/* we have cell indices, now see if they are in bounds */
		if ((0 <= cell_x && cell_x < LongitudeCount()) &&
			(0 <= cell_y && cell_y < LatitudeCount()))
			return m_vData[cell_x].data[cell_y];
		else return -400;
	}
	else
	{
		/* find highest in radius */
		float del, space;
		int xmin, ymin, xmax, ymax;
		int i, j, new_x, new_y;
		short maxalt = -400, val;
		Coord *c;

		if (Level() == 1)
			space = 100.0;
		else if (Level() == 2)
			space = 30.0;
		else space = 100.0;

		del = (float)radius/(float)space;
		xmin = cell_x-(int)del;
		ymin = cell_y-(int)del;
		xmax = cell_x+(int)del;
		ymax = cell_y+(int)del;

		c = MapPostToCoord(cell_x, cell_y);
		c->Print();

		for (i = xmin; i < xmax; i++)
		{
			if (i < 0 || i > LongitudeCount())
				continue;
			for (j = ymin; j < ymax; j++)
			{
				if (j < 0 || j > LatitudeCount())
					continue;
				val = m_vData[i].data[j];
				if (val > maxalt)
				{
					maxalt = val;
					new_x = i;
					new_y = j;
					/*printf("indices: (%d, %d) => (%d, %d)\n", cell_x, cell_y, i, j);*/
				}
			}
		}
		cout<<endl;
		int dx = new_x-cell_x;
		int dy = new_y-cell_y;
		float rng = sqrt((float)((dx*dx)+(dy*dy)))*space;
		cout<<"range: "<<rng<<"m\n\n";
		Coord *c2 = MapPostToCoord(new_x, new_y);
		c2->Print();
		cout<<endl;
		delete c2;
		return maxalt;
	}

	return 0;
}

short DTED::AltitudeBySite(Site *point)
{
	return AltitudeByGeodetic(point->Latitude(), point->Longitude(), SMOOTH_ALTITUDE);
}


int DTED::LatitudeCount()
{
	return (int)m_vData.size();
}

int DTED::LongitudeCount()
{
	if (m_vData.size() > 0)
		return m_vData[0].count;
	else return 0;
}

int DTED::MinAltitude()
{
#if 0
	int x, y, v = 65535;
	int xmax = (int)m_vData.size();
	int ymax = m_vData[0].count;

	for (x = 0; x < xmax; x++)
	{
		for (y = 0; y < ymax; y++)
		{
			if (m_vData[x].data[y] < v)
				v = m_vData[x].data[y];
		}
	}
	return v;
#else
	return m_iMinAltitude;
#endif
}

int DTED::MaxAltitude()
{
#if 0
	int x, y, v = -65535;
	int xmax = (int)m_vData.size();
	int ymax = m_vData[0].count;

	for (x = 0; x < xmax; x++)
	{
		for (y = 0; y < ymax; y++)
		{
			if (m_vData[x].data[y] > v)
				v = m_vData[x].data[y];
		}
	}
	return v;
#else
	return m_iMaxAltitude;
#endif
}

short DTED::Interval()
{
	if (m_sLevel == 1)
		return 100;
	if (m_sLevel == 2)
		return 30;
	return 100;
}

void DTED::CalcWindowSize(int &xsize, int &ysize, float &xrat, float &yrat)
{
	/*int x_max = LongitudeCount();
	int y_max = LatitudeCount();*/
	int xsz = WINDOW_SIZE, ysz = WINDOW_SIZE;
	int xspn, yspn;
	float xn, yn;

	if (m_iPlotXMax != -1)
		xspn = m_iPlotXMax;
	else xspn = LongitudeCount();

	if (m_iPlotXMin != -1)
		xspn -= m_iPlotXMin;

	if (m_iPlotYMax != -1)
		yspn = m_iPlotYMax;
	else yspn = LatitudeCount();

	if (m_iPlotYMin != -1)
		yspn -= m_iPlotYMin;

	xn = (float)xspn/(float)LongitudeCount();
	yn = (float)yspn/(float)LatitudeCount();
	printf("xn = %f, yn = %f\n", xn, yn);

	xsz = (int)(xn*(float)xsz);
	ysz = (int)(yn*(float)ysz);

	xsz *= m_iPlotZoomLevel;
	ysz *= m_iPlotZoomLevel;

	printf("xsz = %d, ysz = %d\n", xsz, ysz);

	xsize = xsz;
	ysize = ysz;

	xrat = xn;
	yrat = yn;
}

int DTED::str2lat(const char *str, int len)
{
	char c1 = tolower(str[0]);
	char c2 = tolower(str[len-1]);

	if (c1 == 'n' || c1 == 's')
	{
		// filevec style (['n'|'s'][deg])
		int val, dir;
		sscanf(str, "%c%d", &c1, &val);
		dir = (tolower(c1) == 'n')?1:-1;
		return val*dir;
	}
	else if (c2 == 'n' || c2 == 's')
	{
		// DTED style (DDDMMSSH)
		int deg, min, sec, dir;
		sscanf(str, "%3d%2d%2d%c", &deg, &min, &sec, &c1);
		dir = (tolower(c1) == 'n')?1:-1;
		return deg*dir;
	}
	else return -65535;
}

int DTED::str2lon(const char *str, int len)
{
	char c1 = tolower(str[0]);
	char c2 = tolower(str[len-1]);

	if (c1 == 'w' || c1 == 'e')
	{
		// filevec style (['w'|'e'][deg])
		int val, dir;
		sscanf(str, "%c%d", &c1, &val);
		dir = (tolower(c1) == 'e')?1:-1;
		return val*dir;
	}
	else if (c2 == 'w' || c2 == 'e')
	{
		// DTED style (DDDMMSSH)
		int deg, min, sec, dir;
		sscanf(str, "%3d%2d%2d%c", &deg, &min, &sec, &c1);
		dir = (tolower(c1) == 'e')?1:-1;
		return deg*dir;
	}
	else return -65535;
}

string DTED::lat2str(int lat)
{
	char buf[32];
	char c;

	c = (lat < 0)?'s':'n';
	lat = abs(lat);
	_snprintf(buf, 32, "%c%d", c, lat);

	return string(buf);
}

string DTED::lon2str(int lon)
{
	char buf[32];
	char c;

	c = (lon < 0)?'w':'e';
	lon = abs(lon);
	_snprintf(buf, 32, "%c%d", c, lon);

	return string(buf);
}

void DTED::MapCoordToPost(double lat, double lon, int &x_index, int &y_index/*, bool interpolate*/)
{
	double reflat = (double)m_Header.uhl.proc.origin_lat;
	double reflon = (double)m_Header.uhl.proc.origin_lon;
	double dlat = lat-reflat;
	double dlon = lon-reflon;

	/* convert delta lat/lon from degrees to 1/10 seconds */
	dlat *= 36000;
	dlon *= 36000;

	/* divide to interval values to get cell indices */
#ifdef _INTERPOLATE_COORD
	if (interpolate == true)
	{
		bool do_n[2] = {true, true};
		unsigned short xalt[2], yalt[2];
		float nxalt, nyalt;
		float xf, yf;

		/* more work if we want more precise altitudes for positions not
		precisely on the post indices */
		xf = dlon/(float)m_Header.uhl.proc.interval_lon;
		yf = dlat/(float)m_Header.uhl.proc.interval_lat;

		x_index = (long)xf;
		y_index = (long)yf;

		xf -= (float)x_index;
		yf -= (float)y_index;

		printf("MapCoordToPost: xf = %f, yf = %f\n", xf, yf);

		/* if xf == 0.0, we are right on the X post
		if yf == 0.0, we are right on the Y post */
		if (xf == 0.0)
			do_n[0] = false;
		if (yf == 0.0)
			do_n[1] = false;

		/* smooth each axis (x first, then y) */
		if (do_n[0] == true)
		{
			/* get altitudes for bounding x-posts */

		}

		if (do_n[1] == true)
		{
			/* get altitudes for bounding y-posts */
		}

	}
	else
#endif
	{
		x_index = (long)(dlon/(float)m_Header.uhl.proc.interval_lon);
		y_index = (long)(dlat/(float)m_Header.uhl.proc.interval_lat);
	}
}

void DTED::MapPostToGraphic(int &x_index, int &y_index, int xmin, int ymin)
{
	int xoffo, yoffo, xoff, yoff;
	int x_max = LongitudeCount();
	int y_max = LatitudeCount();

	if (PlotOriginLon != "" && PlotOriginLat != "")
	{
		xoffo = str2lon(PlotOriginLon.c_str(), (int)PlotOriginLon.length());
		yoffo = str2lat(PlotOriginLat.c_str(), (int)PlotOriginLat.length());
		xoff = m_Header.uhl.proc.origin_lon;
		yoff = m_Header.uhl.proc.origin_lat;

		xoff = (xoff-xoffo)*x_max;
		yoff = (yoff-yoffo)*y_max;
	}
	else
	{
		xoff = 0;
		yoff = 0;
	}

	x_index += xoff;
	y_index += yoff;

	/* Compensate for tight bounds */
#if 0
	if (xmin != -1)
		x_index -= xmin;

	if (ymin != -1)
		y_index -= ymin;
#endif
}

Coord *DTED::MapPostToCoord(int x_index, int y_index)
{
	int x_step, y_step;
	double lat, lon;
	Coord *c;

	/* calculate step by index*interval for longitude */
	x_step = x_index*m_Header.uhl.proc.interval_lon;

	/* calculate step by index*interval for latitude */
	y_step = y_index*m_Header.uhl.proc.interval_lat;

	/* divide by 36000 to get fractions of a degree */
	lat = (float)y_step/36000.0;
	lon = (float)x_step/36000.0;

	/* add longitude step to origin longitude to get true longitude */
	lon += m_Header.uhl.proc.origin_lon;

	/* add latitude step to origin latitude to get true latitude */
	lat += m_Header.uhl.proc.origin_lat;

	/* set and return a Coord */
	c = new Coord(lat, lon, AltitudeByPost(x_index, y_index), 0.0);

	return c;
}

#ifdef WIN32
string DTED::FindFile(string dpath, string filespec)
{
	struct _finddata_t fileinfo;
	string fname, path;
	bool found;
	intptr_t r;
	int v;

	path = dpath+filespec+".dt*";
	r = _findfirst(path.c_str(), &fileinfo);
	if (r == -1)
	{
		perror("_findfirst: ");
		return "";
	}

	found = false;
	fname = "";
	v = 0;

	while (!found && v == 0)
	{
		char *p;

		if (strlen(fileinfo.name) < 4)
			goto _next_file;

		if (strncmp(fileinfo.name+(strlen(fileinfo.name)-4), ".dt", 3) != 0)
			goto _next_file;

		p = strchr(fileinfo.name, '.');
		if (p == NULL)
			goto _next_file;

		fname = fileinfo.name;
		found = true;
#ifdef _FF_DEBUG
		printf("findfile: %d %s\n", fileinfo.attrib, fileinfo.name);
#endif

_next_file:
		v = _findnext(r, &fileinfo);
	}

	_findclose(r);

	return fname;
}
#else
string DTED::FindFile(string dpath, string filespec)
{
	DIR *dp;
	struct dirent *de;
	string fname;
	bool found;

	dp = opendir(dpath.c_str());
	if (dp == NULL)
	{
		perror("findfile: ");
		return "";
	}

	found = false;
	fname = "";

	while (!found && (de = readdir(dp)) != NULL)
	{
		char *p;

		if (de->d_type == DT_FIFO ||
			de->d_type == DT_CHR ||
			de->d_type == DT_DIR ||
			de->d_type == DT_BLK ||
			de->d_type == DT_SOCK ||
			de->d_type == DT_WHT)
			continue;

		if (strlen(de->d_name) < 4)
			continue;

		if (strncmp(de->d_name+(strlen(de->d_name)-4), ".dt", 3) != 0)
			continue;

		p = strchr(de->d_name, '.');
		if (p == NULL)
			continue;

		if (!strncmp(de->d_name, filespec.c_str(), p-de->d_name))
		{
			fname = de->d_name;
			found = true;
#ifdef _FF_DEBUG
			printf("findfile: %d %s\n", de->d_type, de->d_name);
#endif
		}
	}

	closedir(dp);

	return fname;
}
#endif

void DTED::BuildFileList(double xmin, double ymin, int xcnt, int ycnt,
						 string path, vector<string> &list)
{
	char dir_prefix, file_prefix, delim;
	int i, j;
	string p;
	char pbuf[64], fbuf[64];
	static string msg;

#ifdef WIN32
	delim = '\\';
#else
	delim = '/';
#endif

	/* clear file list */
	list.clear();

	/* set directory prefix based on longitude */
	if (xmin < 0.0)
		dir_prefix = 'w';
	else dir_prefix = 'e';

	/* set file prefix based on latitude */
	if (ymin < 0.0)
		file_prefix = 's';
	else file_prefix = 'n';

	/* make sure the path has a trailing / */
	if (path[path.length()-1] != delim)
		path += delim;

	/* for each longitude ... */
	for (i = 0; i < xcnt; i++)
	{
		/* for each latitude ... */
		for (j = 0; j < ycnt; j++)
		{
			/* format directory string */
			sprintf(pbuf, "%c%d%c", dir_prefix, abs((int)xmin+i), delim);

			/* format file string */
			sprintf(fbuf, "%c%d", file_prefix, abs((int)ymin+j));

#ifdef _FF_DEBUG
			printf("pbuf = %s, fbuf = %s\n", pbuf, fbuf);
#endif
			/* try to find the file */
			p = FindFile(path+string(pbuf), string(fbuf));

			/* an empty string means failure */
			if (p == "")
			{
				msg = "could not find a matching file: path="+path+", pbuf="+
					pbuf+", fbuf="+fbuf;
				cout<<msg<<endl;
			}
			else
			{
				/* build fully qualified path name */
				p = path+string(pbuf)+p;
#ifdef _FF_DEBUG
				cout<<"found file: "<<p<<endl;
#endif
				/* add file path to list */
				list.push_back(p);
			}
		}
	}
}

void DTED::MapCoordToGraphic(double lat, double lon, int &x_index, int &y_index)
{
	double reflat = (double)m_Header.uhl.proc.origin_lat;
	double reflon = (double)m_Header.uhl.proc.origin_lon;
	double dlat = lat-reflat;
	double dlon = lon-reflon;

	/* convert delta lat/lon from degrees to 1/10 seconds */
	dlat *= 36000;
	dlon *= 36000;

	/* divide to interval values to get cell indices */
	x_index = (long)(dlon/(float)m_Header.uhl.proc.interval_lon);
	y_index = (long)(dlat/(float)m_Header.uhl.proc.interval_lat);

	int xoff = LongitudeCount()*m_iLonOffset;
	int yoff = LatitudeCount()*m_iLatOffset;

	x_index += xoff;
	y_index += yoff;
}
