#ifdef _XPLOTTER
#include <plotter.h>
#endif
#include "dtedgrid.h"

#include "dted.h"
#include "fastpath.h"
#include "site.h"

using namespace std;

#define SMOOTH_ALTITUDE false

DTEDGrid *DTEDGrid::gPtr = NULL;

DTEDGrid::DTEDGrid(int width, int height, double origin_lat, double origin_lon)
{
	m_iWidth = width;
	m_iHeight = height;
	m_dOriginLat = origin_lat;
	m_dOriginLon = origin_lon;
	m_iLevel = -1;
    m_pStatusCB = NULL;
    m_pStatusCB_Data = NULL;

	m_bAllocated = _alloc();
	gPtr = this;

	DTED::PlotOriginLat = DTED::lat2str((long)origin_lat);
	DTED::PlotOriginLon = DTED::lon2str((long)origin_lon);
}

DTEDGrid::~DTEDGrid()
{
	m_bAllocated = !_dealloc();
}


bool DTEDGrid::AddCell(DTED *cell, int &x_index, int &y_index)
{
	double cell_lat, cell_lon, dlat, dlon;

	if (m_iWidth == 0 || m_iHeight == 0 || cell == NULL)
		return false;

	if (m_iLevel == -1)
	{
		m_iLevel = cell->Level();
	}
	else
	{
		if (cell->Level() != m_iLevel)
			return false;
	}

	cell_lat = cell->OriginLatVal();
	cell_lon = cell->OriginLonVal();
	dlat = cell_lat-m_dOriginLat;
	dlon = cell_lon-m_dOriginLon;

	/* using cell lat/lon, figure out where in the grid this cell goes */
	/* --------------------------------------------------------------- */
	/* determine cell index */
	x_index = (long)dlon;
	y_index = (long)dlat;

	/*cout<<"\tgrid location ("<<x_index<<", "<<y_index<<")\n";*/
	if (x_index >= m_iWidth || y_index >= m_iHeight)
		return false;

	/* if the current grid value is non-NULL, print a warning and free it */
	if (m_ppGrid[x_index][y_index] != NULL)
		delete m_ppGrid[x_index][y_index];

	/* place cell into the grid location */
	m_ppGrid[x_index][y_index] = cell;

	/* tell cell which position it is in */
	cell->LatOffset(y_index);
	cell->LonOffset(x_index);

	return true;
}

bool DTEDGrid::DelCell(int x_index, int y_index)
{
	/* check if indices are in bounds */
	if (0 <= x_index && x_index < m_iWidth &&
		0 <= y_index && y_index < m_iHeight)
	{
		/* check if the cell is currently allocated */
		if (m_ppGrid[x_index][y_index] == NULL)
			return true;

		/* remove indicated cell */
		delete m_ppGrid[x_index][y_index];
		m_ppGrid[x_index][y_index] = NULL;

		/* return success */
		return true;
	}

	/* return failed */
	return false;
}

DTED *DTEDGrid::Cell(int x, int y)
{
	if ((0 <= x && x < m_iWidth) &&
		(0 <= y && y < m_iHeight))
		return m_ppGrid[x][y];
	else return NULL;
}

void DTEDGrid::Cell(DTED *cell, int x, int y)
{
	if ((0 <= x && x < m_iWidth) &&
		(0 <= y && y < m_iHeight))
	{
		if (m_ppGrid[x][y] != NULL)
			delete m_ppGrid[x][y];
		m_ppGrid[x][y] = cell;
	}
}

short DTEDGrid::AltitudeByPost(int x_index, int y_index)
{
	DTED *d;

	d = MapGraphicToCellPtr(&x_index, &y_index);
	if (d != NULL)
	{
		/*printf("x = %d, y = %d, lat = %d, lon = %d\n",
		x_index, y_index, d->LatOffset(), d->LonOffset());*/
		return d->AltitudeByPost(x_index, y_index);
	}
	else return -400;
}

short DTEDGrid::AltitudeByGeodetic(double lat, double lon, bool interpolate)
{
	int x_index, y_index;
	short alt;

	/* using the origin lat/lon for the grid, calculate the cell x/y for the lookup */
	MapCoordToCell(lat, lon, &x_index, &y_index);

	/* call DTED::AltitudeByGeodetic() on the chosen cell with the given lat/lon */
	if ((0 <= x_index && x_index < m_iWidth) &&
		(0 <= y_index && y_index < m_iHeight))
		alt = m_ppGrid[x_index][y_index]->AltitudeByGeodetic(lat, lon, interpolate, 0);
	else alt = -400;

	/* return the value */
	return alt;
}

int DTEDGrid::LatitudeCount()
{
	return DTED::PointCount[m_iLevel]*m_iHeight;
} 

int DTEDGrid::LongitudeCount()
{
	return DTED::PointCount[m_iLevel]*m_iWidth;
} 

int DTEDGrid::BoundWidth()
{
	return LongitudeCount()-m_iWidth;
}

int DTEDGrid::BoundHeight()
{
	return LatitudeCount()-m_iHeight;
}

int DTEDGrid::MinAltitude()
{
	int x, y, val = 65535;
	int alt;

	if (m_iHeight == 0 || m_iWidth == 0)
		return 1;

	for (x = 0; x < m_iWidth; x++)
	{
		for (y = 0; y < m_iHeight; y++)
		{
			if (m_ppGrid[x][y] != NULL)
			{
				alt = m_ppGrid[x][y]->MinAltitude();
				if (alt < val)
					val = alt;
			}
		}
	}

	return val;
}

int DTEDGrid::MaxAltitude()
{
	int x, y, val = -65535;
	int alt;

	if (m_iHeight == 0 || m_iWidth == 0)
		return 1;

	for (x = 0; x < m_iWidth; x++)
	{
		for (y = 0; y < m_iHeight; y++)
		{
			if (m_ppGrid[x][y] != NULL)
			{
				alt = m_ppGrid[x][y]->MaxAltitude();
				if (alt > val)
					val = alt;
			}
		}
	}

	return val;
}

int DTEDGrid::LatCountPerCell()
{
	if (m_iLevel == -1)
		return 0;

	return DTED::PointCount[m_iLevel];
}

int DTEDGrid::LonCountPerCell()
{
	if (m_iLevel == -1)
		return 0;

	return DTED::PointCount[m_iLevel];
}

DTED *DTEDGrid::_dtedlookupfunc(double lat, double lon)
{
	int x_index, y_index;
	static int gwidth = gPtr->Width();
	static int gheight = gPtr->Height();

	/* using the origin lat/lon for the grid, calculate the cell x/y for the lookup */
	x_index = (long)(lon-gPtr->OriginLon());
	y_index = (long)(lat-gPtr->OriginLat());

#ifdef _DTEDGRID_DEBUG
	printf("DEBUG::_dtedlookupfunc(%3.2f, %3.2f): x_index = %d, y_index = %d, " \
		"dlat = %f, dlon = %f, ",
		lat, lon, x_index, y_index,
		lat-gPtr->OriginLat(),lon-gPtr->OriginLon());
#endif
	if ((0 <= x_index && x_index < gwidth) &&
		(0 <= y_index && y_index < gheight))
		return gPtr->Cell(x_index, y_index);
	else return NULL;
}

FastPath *DTEDGrid::GeneratePath(Site *src, Site *dest, float inc, bool interpalt)
{
	FastPath *path;

	path = new FastPath;

	cout<<"Source:\n";
	src->Print();
	cout<<"\nDestination:\n";
	dest->Print();

	path->CreatePath2(src, dest, (DTED*)NULL, _dtedlookupfunc, true, inc, interpalt);
	cout<<"\nPath length: "<<path->Size()<<endl;

	return path;
}

/* map (lon, lat) as coords in grid to which DTED cell (lon, lat) are in */
DTED *DTEDGrid::MapCoordToCell(double lat, double lon, int *x_index, int *y_index)
{
	double dlat = lat-m_dOriginLat;
	double dlon = lon-m_dOriginLon;

	//x_index = (long)(dlon/(float)m_iWidth);
	//y_index = (long)(dlat/(float)m_iHeight);
	*x_index = (long)dlon;
	*y_index = (long)dlat;

	if (0 <= *x_index && *x_index < m_iWidth && 0 <= *y_index && *y_index < m_iHeight)
		return m_ppGrid[*x_index][*y_index];
	else return NULL;
}

/* map (x, y) as posts in grid to which DTED cell (x, y) are in */
DTED *DTEDGrid::MapGraphicToCellPtr(int *x_index, int *y_index)
{
	int xind, yind;

	if (m_iLevel == -1)
		return NULL;

	int ocount = DTED::PointCount[m_iLevel];
	int acount = ocount;

	xind = *x_index/ocount;
	yind = *y_index/acount;

	*x_index -= (xind*ocount);
	*y_index -= (yind*acount);

	if (0 <= xind && xind < m_iWidth &&
		0 <= yind && yind < m_iHeight)
		return m_ppGrid[xind][yind];

	return NULL;
}

void DTEDGrid::MapCoordToGraphic(double lat, double lon, int *x_index, int *y_index)
{
	int xc, yc;
	DTED *d;

	d = MapCoordToCell(lat, lon, &xc, &yc);
	if (d != NULL)
	{
		/* Map lat/lon coordinates to post indices in the cell */
		d->MapCoordToPost(lat, lon, *x_index, *y_index/*, SMOOTH_ALTITUDE*/);

		/* Map post indices to graphic coordinates */
		d->MapPostToGraphic(*x_index, *y_index, -1, -1);
	}
	else
	{
		*x_index = -1;
		*y_index = -1;
	}
}

Coord *DTEDGrid::MapGraphicToCoord(int x_index, int y_index)
{
	int xc = x_index, yc = y_index;
	DTED *d;
	Coord *c;

	d = MapGraphicToCellPtr(&xc, &yc);
	if (d != NULL)
	{
		/* map graphic coordinates to post indices */
#if 0
		x_index -= (d->LongitudeCount() * xc);
		y_index -= (d->LatitudeCount() * yc);
#else
		x_index -= (d->LongitudeCount() * d->LonOffset());
		y_index -= (d->LatitudeCount() * d->LatOffset());
#endif

		/* map post indices to lat/lon coordinates */
		c = d->MapPostToCoord(x_index, y_index);

		return c;
	}
	else return NULL;
}


bool DTEDGrid::_alloc()
{
	int i, j;

	m_ppGrid = new DTED**[m_iWidth];
	for (i = 0; i < m_iWidth; i++)
	{
		m_ppGrid[i] = new DTED*[m_iHeight];

		for (j = 0; j < m_iHeight; j++)
			m_ppGrid[i][j] = NULL;
	}

	return true;
}

bool DTEDGrid::_dealloc()
{
	int i, j;

	for (i = 0; i < m_iWidth; i++)
	{
		for (j = 0; j < m_iHeight; j++)
		{
			delete m_ppGrid[i][j];
			m_ppGrid[i][j] = NULL;
		}
		delete [] m_ppGrid[i];
	}
	delete [] m_ppGrid;

	m_ppGrid = NULL;

	return true;
}

DTEDGrid *DTEDGrid::Load(double bxmin, double bymin, double bxmax,
						 double bymax, string path, bool verbose)
{
	int i;
	int cell_x_index, cell_y_index;
	int cell_x_cnt, cell_y_cnt;
	DTEDGrid *grid;
	DTED *d;
	vector<string> list;

	cell_x_cnt = abs((int)(bxmax-bxmin));
	cell_y_cnt = abs((int)(bymax-bymin));
	if (verbose)
		printf("bounds: min=(%d, %d), max=(%d, %d), cnt=(%d, %d)\n",
		(int)bxmin, (int)bymin,
		(int)bxmax, (int)bymax,
		cell_x_cnt, cell_y_cnt);

	grid = new DTEDGrid(cell_x_cnt, cell_y_cnt, bymin, bxmin);
	DTED::BuildFileList(bxmin, bymin, cell_x_cnt, cell_y_cnt, path, list);

	for (i = 0; i < (int)list.size(); i++)
	{
		d = new DTED(list[i]);

		if (d->Loaded())
		{
			grid->AddCell(d, cell_x_index, cell_y_index);
			if (verbose)
			{
				printf("added DTED cell %s at (%d, %d)\n", list[i].c_str(), 
						 cell_x_index, cell_y_index);
			}
		}
		else
		{
			cout<<"DTEDGrid(): Failed to load DTED \""<<list[i]<<"\"\n";
			delete grid;
			return NULL;
		}
	}

	return grid;
}

void DTEDGrid::CalcWindowSize(int &xsize, int &ysize, float &xrat, float &yrat)
{
	/*int x_max = LongitudeCount();
	int y_max = LatitudeCount();*/
	int xsz = WINDOW_SIZE, ysz = WINDOW_SIZE;
	int xspn, yspn;

	xspn = BoundWidth();
	yspn = BoundHeight();

	xsize = xspn;
	ysize = yspn;

	xrat = 1.0;
	yrat = 1.0;
}

void DTEDGrid::SetStatusCallback(StringCallback p, void *data)
{
    m_pStatusCB = p;
    m_pStatusCB_Data = data;
}

DTEDGrid::StringCallback DTEDGrid::GetStatusCallback(void **data)
{
	*data = m_pStatusCB_Data;
	return m_pStatusCB;
}

