#include "fastpath.h"

#include "site.h"
#include "dted.h"
#include "greatcircle.h"

#define _CREATE_PATH2_USE_FRACTION

bool FastPath::Debug = false;

FastPath::FastPath()
{
  m_ipAltitude = NULL;
  m_fpDistance = NULL;
  m_pSites = NULL;
  m_iSize = 0;
  m_iCapacity = 0;
  m_fDeltaDistance = 0.0;

#ifdef _GC_DEBUG
  m_pFilePtr = fopen("fp.txt", "w+");
#endif
}

FastPath::~FastPath()
{
  m_iSize = 0;
  m_iCapacity = 0;
  
  delete [] m_ipAltitude;
  m_ipAltitude = NULL;
  delete [] m_fpDistance;
  m_fpDistance = NULL;
  delete [] m_pSites;
  m_pSites = NULL;

#ifdef _GC_DEBUG
  delete m_pFilePtr; 
  m_pFilePtr = NULL;
#endif
}

/* Adds a new coordinate to the list.  If 'dist' is non-zero,
   it is assumed to be in meters */
int FastPath::_addpoint(int alt, float dist)
{
  /* check to see if we have enough space in the arrays */
  if (m_iSize == m_iCapacity)
    {
      int *atmp;
      float *dtmp;
      Site **stmp;
      int newsz;

      /* new size is m_iCapacity*2 */
      newsz = m_iCapacity*2;

      /* create new arrays of the correct size */
      atmp = new int[newsz];
      dtmp = new float[newsz];
      stmp = new Site*[newsz];

      /* copy altitude data to new array */
      memcpy(atmp, m_ipAltitude, m_iSize*sizeof(int));

      /* copy distance data to new array */
      memcpy(dtmp, m_fpDistance, m_iSize*sizeof(float));

      /* copy site data to new array */
      memcpy(stmp, m_pSites, m_iSize*sizeof(Site*));

      /* free old arrays */
      delete [] m_ipAltitude;
      delete [] m_fpDistance;
      delete [] m_pSites;

      /* update pointers and sizes */
      m_ipAltitude = atmp;
      m_fpDistance = dtmp;
      m_pSites = stmp;
      m_iCapacity = newsz;

#ifdef _GC_DEBUG
      fprintf(m_pFilePtr, "_addpoint: capacity increasing to %d\n", m_iCapacity);
      fflush(m_pFilePtr);
#endif
    }

  /* place altitude and distance into the next place in the arrays */
  m_ipAltitude[m_iSize] = alt;
  m_fpDistance[m_iSize] = dist;
  m_pSites[m_iSize] = NULL;

  /* increment size */
  m_iSize++;

  /* return index of newly added data */
  return m_iSize-1;
}

int FastPath::Altitude(int index)
{
  if (0 <= index && index < m_iSize)
    return m_ipAltitude[index];
  else return -6000;
}

float FastPath::Distance(int index)
{
  if (0 <= index && index < m_iSize)
    return m_fpDistance[index];
  else return -1.0;
}

Site *FastPath::Sites(int index)
{
  if (0 <= index && index < m_iSize)
    return m_pSites[index];
  else return NULL;
}

/* This function generates a sequence of latitude and longitude positions
   between a source location and a destination along a great circle path,
   and stores elevation and distance information for points along that
   path in the "path" structure for later use. */
bool FastPath::CreatePath1(Site *src, Site *dest, DTED *cell, DTEDLookupCallback callBack,
			   bool include_lastpos, float inc, bool interpalt)
{
  double azimuth, distance, lat1, lon1, beta;
  double den, num, lat2, lon2, total_distance;
  double altstep = 0.0, diststep = 0.0;
  int x0, y0, x1, y1;
  Site tempsite;
  DTED *lookup = NULL;

  x0=0;
  y0=0;

  /* latitude in radians */
  lat1 = src->Latitude()*DEG2RAD;
  /* longitude in radians */
  lon1 = src->Longitude()*DEG2RAD;
  /* azimuth in radians, math coordinate frame */
  azimuth = (360.0-src->Azimuth(*dest))*DEG2RAD;
  /* distance in nautical miles */
  total_distance = *src-*dest;

  /* interpolate altitudes, i.e., for flight path */
  if (interpalt == true)
    {
      altstep = dest->Altitude()-src->Altitude();
      diststep = total_distance/inc;
      if (diststep != 0.0)
	altstep /= diststep;
    }

  /* clear arrays */
  m_iSize = 0;
  m_iCapacity = 0;
  if (m_ipAltitude != NULL)
    delete [] m_ipAltitude;
  if (m_fpDistance != NULL)
    delete [] m_fpDistance;
  if (m_pSites != NULL)
    delete [] m_pSites;

  /* pre-calculate approximate array size */
  m_iCapacity = (int)(total_distance/inc)+2;
  
  /* allocate new arrays */
  m_ipAltitude = new int[m_iCapacity];
  m_fpDistance = new float[m_iCapacity];
  m_pSites = new Site*[m_iCapacity];
  
  for (distance = 0; distance <= total_distance; distance += inc)
    {
      /*beta = distance/3959.0;*/
      beta = distance/3436.029;
      lat2 = asin(sin(lat1)*cos(beta)+cos(azimuth)*sin(beta)*cos(lat1));
      num = cos(beta)-(sin(lat1)*sin(lat2));
      den = cos(lat1)*cos(lat2);

      if (azimuth == 0.0 && (beta > (M_PI/2.0)-lat1))
	{
	  lon2 = lon1+PI;
	}
      else if (azimuth == HALFPI && (beta > (M_PI/2.0)+lat1))
	{
	  lon2 = lon1+PI;
	}
      else if (fabs(num/den) > 1.0)
	{
	  lon2 = lon1;
	}
      else
	{
	  if ((M_PI-azimuth) >= 0.0)
	    {
	      lon2 = lon1-_arccos(num, den);
	    }
	  else
	    {
	      lon2 = lon1+_arccos(num, den);
	    }
	}

      /* This while() test broke the implementation for a while; going from 118W => 117W would
	 produce longitudes of 252E. */
      /*while (lon2 < 0.0)
	lon2 += (M_PI*2.0);*/
      
      while (lon2 > (M_PI*2.0))
	{
	  lon2 -= (M_PI*2.0);
	}
 
      lat2 = lat2/DEG2RAD;
      lon2 = lon2/DEG2RAD;

      x1 = (int)(1199.0*(lat2-floor(lat2)));
      y1 = (int)(1199.0*(lon2-floor(lon2)));

      if (x1 != x0 || y1 != y0)
	{
	  tempsite.Latitude(lat2);
	  tempsite.Longitude(lon2);

	  if (callBack != NULL)
	    lookup = callBack(lat2, lon2);
	  else lookup = cell;
	  
	  if (lookup == NULL && cell != NULL)
	    lookup = cell;
	  
	  if (lookup != NULL)
	    {
	      double alt;
	      
	      if (interpalt)
		alt = src->Altitude()+((distance/inc)*altstep);
	      else alt = lookup->AltitudeByGeodetic(lat2, lon2, false, 0);

	      _addpoint((int)alt, (float)GC_NM_TO_METERS(distance));
	    }
	}

      x0 = x1;
      y0 = y1;
    }

  if (include_lastpos == true)
    {
      /* Make sure exact dest point is recorded at path.length-1 */
      if (callBack != NULL)
	lookup = callBack(dest->Latitude(), dest->Longitude());
      else lookup = cell;
      
      if (lookup == NULL && cell != NULL)
	lookup = cell;
      
      if (lookup != NULL)
	{
	  double alt;
	  
	  if (interpalt)
	    alt = dest->Altitude();
	  else alt = lookup->AltitudeByGeodetic(dest->Latitude(), dest->Longitude(), false, 0);
	  
	  _addpoint((int)alt, (float)GC_NM_TO_METERS(total_distance));
	}
    }
  
  return true;
}

/* This function generates a sequence of latitude and longitude positions
   between a source location and a destination along a great circle path,
   and stores elevation and distance information for points along that
   path in the "path" structure for later use. */
bool FastPath::CreatePath2(Site *src, Site *dest, DTED *cell, DTEDLookupCallback cbDted,
			   bool include_lastpos, float inc, bool interpalt)
{
  double distance, lat1, lon1;
  double lat2, lon2, total_distance;
  double latx, lonx;
  double altstep = 0.0, diststep = 0.0;
  int cnt;
  DTED *lookup = NULL;
  bool ret;

  cnt = 0;

  /* distance in nautical miles */
  total_distance = *src-*dest;

  /* interpolate altitudes, i.e., for flight path */
  if (interpalt == true)
    {
      altstep = dest->Altitude()-src->Altitude();
      diststep = total_distance/inc;
      if (diststep != 0.0)
	altstep /= diststep;
    }

  /* latitude in radians */
  lat1 = src->Latitude()*DEG2RAD;
  /* longitude in radians */
  lon1 = src->Longitude()*DEG2RAD;
  /* latitude in radians */
  lat2 = dest->Latitude()*DEG2RAD;
  /* longitude in radians */
  lon2 = dest->Longitude()*DEG2RAD;

  if (Debug)
    printf("CreatePath2: total_distance = %f nm\n", total_distance);

  /* clear arrays */
  m_iSize = 0;
  m_iCapacity = 0;
  if (m_ipAltitude != NULL)
    delete [] m_ipAltitude;
  if (m_fpDistance != NULL)
    delete [] m_fpDistance;
  if (m_pSites != NULL)
    delete [] m_pSites;

  /* pre-calculate approximate array size */
  m_iCapacity = (int)(total_distance/inc)+2;
  
  /* allocate new arrays */
  m_ipAltitude = new int[m_iCapacity];
  m_fpDistance = new float[m_iCapacity];
  m_pSites = new Site*[m_iCapacity];

  /* for each increment ... */
  for (distance = 0; distance <= total_distance; distance += inc)
    {
      /* position returned (latx/lonx) is in radians */
#ifdef _CREATE_PATH2_USE_FRACTION
      ret = GreatCircle::IntermediateByFraction(lat1, lon1, lat2, lon2,
						&latx, &lonx,
						total_distance,
						(float)(distance/total_distance));
#else
      ret = GreatCircle::IntermediateByDistance(lat1, lon1, lat2, lon2,
						&latx, &lonx,
						distance, total_distance);
#endif

      if (ret == false)
	{
	  printf("GreatCircle::IntermediateByDistance() failed on distance %f\n", distance);
	  continue;
	}

      latx = GC_RADIANS_TO_DEGREES(latx);
      lonx = GC_RADIANS_TO_DEGREES(lonx);

      if (cbDted != NULL)
	lookup = cbDted(latx, lonx);
      else lookup = cell;
	  
      if (lookup == NULL && cell != NULL)
	lookup = cell;
	  
      if (lookup != NULL)
	{
	  double alt;
	  
	  if (interpalt)
	    alt = src->Altitude()+((distance/inc)*altstep);
	  else alt = lookup->AltitudeByGeodetic(latx, lonx, false, 0);

	  _addpoint((int)alt, (float)GreatCircle::NmToMeters(distance));
	  m_pSites[m_iSize-1] = new Site("", latx, lonx, alt, 0);
	}
      cnt++;
    }

  if (include_lastpos == true)
    {
      /* Make sure exact dest point is recorded at path.length-1 */
      if (cbDted != NULL)
	lookup = cbDted(dest->Latitude(), dest->Longitude());
      else lookup = cell;
      
      if (lookup == NULL && cell != NULL)
	lookup = cell;
      
      if (lookup != NULL)
	{
	  float alt;
	  
	  if (interpalt)
	    alt = (float)dest->Altitude();
	  else alt = lookup->AltitudeByGeodetic(dest->Latitude(), dest->Longitude(), false, 0);

	  _addpoint((int)alt, (float)GC_NM_TO_METERS(total_distance));
	  m_pSites[m_iSize-1] = new Site("", dest->Latitude(), dest->Longitude(), alt, 0);
	}
    }

  /* 'inc' is in nautical miles, so convert to meters */
  m_fDeltaDistance = (float)GC_NM_TO_METERS(inc);

  if (Debug)
    printf("CreatePath2: total count = %d\n", m_iSize);
  
  return true;
}

bool FastPath::SaveToFile(char *fname)
{
  int i;
  ofstream op(fname);
  
  for (i = 0; i < m_iSize; i++)
    op<<i<<"\t"<<m_fpDistance[i]<<"\t"<<m_ipAltitude[i]<<endl;
  op.close();

  return true;
}

/* This function implements the arc cosine function,
   returning a value between 0 and M_PI*2. */
double FastPath::_arccos(double x, double y)
{
  double result = 0.0;

  if (y > 0.0)
    result = acos(x/y);

  if (y < 0.0)
    result = PI+acos(x/y);

  return result;
}

int FastPath::TravelLOS(Site *src, Site *dest, int &obsalt)
{
  int i;
  long index, pathsize;
  bool block;
  double distance, tx_alt, rx_alt;
  int maxalt;
  double xmtr_angle, test_angle, test_alt;

  index = -1;
  pathsize = Size();
  obsalt = -6000;
  maxalt = -6000;
  tx_alt = src->Altitude()+src->Antenna();
  rx_alt = dest->Altitude()+dest->Antenna();
  distance = Distance(pathsize-1);

  if (Debug)
    {
      printf("tx_alt   = %f\n", tx_alt);
      printf("rx_alt   = %f\n", rx_alt);
      printf("distance = %f m\n", distance);
    }

  xmtr_angle = atan2(rx_alt-tx_alt, distance);

  if (Debug)
    printf("xmtr_angle = %f\n", xmtr_angle);

  /* Determine the elevation angle of each discrete location
     along the path between the receiver and transmitter. */
  for (i = 1, block = false; i < pathsize && block == false; i++)
    {
      test_alt = Altitude(i);
      distance = Distance(i);

      if (Debug)
	{
	  printf("i = %d\n", i);
	  printf("\ttest_alt = %f\n", test_alt);
	  printf("\tdistance = %f m\n", distance);
	}
      
      /* Determine the elevation of the test point as seen from
	 the location of the receiver */
      test_angle = atan2(test_alt-tx_alt, distance);

      if (Debug)
	{
	  printf("\ttest_angle = %f\n", test_angle);
	  printf("\tx<t = %d\n", xmtr_angle<test_angle);
	}

      if (xmtr_angle < test_angle)
	{
	  block = true;
	  index = i;

#if 0
	  while (xmtr_angle < test_angle)
	    {
	      test_alt -= 1.0;
	      test_angle = atan2(test_alt-tx_alt, distance);
	    }
	  obsalt = (int)test_alt;
	  printf("\tadjusting alt = %3.2f, cos(angle) = %f, cos(xmtr) = %f\n",
		 test_alt, test_angle, xmtr_angle);
#endif
	}
      
      if (test_alt > maxalt)
	maxalt = (int)test_alt;
    }
  
  obsalt = maxalt;

  return index;
}

float FastPath::FindMaxAngle(Site *src, Site *dest, int &atidx, int &atalt)
{
  int i;
  long index, pathsize;
  double distance, tx_alt, test_angle;
  double max_angle;
  int test_alt;

  max_angle = -90;
  index = -1;
  pathsize = Size();
  atidx = -1;
  atalt = -6000;
  tx_alt = src->Altitude()+src->Antenna();

  if (Debug)
    {
      printf("tx_alt   = %f\n", tx_alt);
      printf("max_angle = %f\n", max_angle);
    }

  /* Determine the elevation angle of each discrete location
     along the path between the receiver and transmitter. */
  for (i = 1; i < pathsize; i++)
    {
      test_alt = Altitude(i);
      distance = Distance(i);

      if (Debug)
	{
	  printf("i = %d\n", i);
	  printf("\ttest_alt = %d\n", test_alt);
	  printf("\tdistance = %f m\n", distance);
	}
      
      /* Determine the elevation of the test point as seen from
	 the location of the receiver */
      test_angle = atan2(test_alt-tx_alt, distance);

      if (Debug)
	{
	  printf("\ttest_angle = %f\n", test_angle);
	  printf("\tmax<test = %d\n", max_angle<test_angle);
	}

      if (test_angle > max_angle)
	{
	  max_angle = test_angle;
	  atidx = i;
	  atalt = test_alt;
	  
	  if (Debug)
	    {
	      printf("\tmax_angle = %f\n", max_angle);
	      printf("\taltitude = %d\n", atalt);
	    }
	}      
    }

  max_angle = GC_RADIANS_TO_DEGREES(max_angle);
  
  return (float)max_angle;
}
