#include "greatcircle.h"
#include <cmath>

/* Calculate the distance from lat1/lon1 to lat2/lon2.
asRadians controls the units of the returned
distance.

Parameter       Direction     Units
-----------------------------------
lat1               ->          rad
lon1               ->          rad
lat2               ->          rad
lon2               ->          rad
asRadians          ->          bool
<return>           ->          [asRadians -> radians
| -> nm ]
*/
double GreatCircle::Distance(double lat1, double lon1,
							 double lat2, double lon2,
							 bool asRadians)
{
	double dlat;
	double dlon;
	double sin_dlat_2;
	double sin_dlon_2;
	double sin_dlat_p2;
	double sin_dlon_p2;
	double sum_terms;
	double sqrt_sum;
	double asin_sqrt;
	double d_rad;

	dlat = lat1-lat2;
	dlon = lon1-lon2;
	sin_dlat_2 = sin(dlat/2.0);
	sin_dlon_2 = sin(dlon/2.0);
	sin_dlat_p2 = sin_dlat_2*sin_dlat_2;
	sin_dlon_p2 = sin_dlon_2*sin_dlon_2;
	sum_terms = sin_dlat_p2+cos(lat1)*cos(lat2)*sin_dlon_p2;
	sqrt_sum = sqrt(sum_terms);
	asin_sqrt = asin(sqrt_sum);

	d_rad = 2.0*asin_sqrt;

	if (!asRadians)
		return GC_RADIANS_TO_NM(d_rad);
	else return d_rad;
}


/* Calculate the bearing from lat1/lon1 to lat2/lon2.
asRadians controls the units of the returned
bearing.

Parameter       Direction     Units
-----------------------------------
lat1               ->          rad
lon1               ->          rad
lat2               ->          rad
lon2               ->          rad
asRadians          ->          bool
<return>           ->          [asRadians -> radians
| -> degrees ]
*/
double GreatCircle::Radial(double lat1, double lon1,
						   double lat2, double lon2,
						   bool asRadians)
{
	double d;
	double base;
	double c_rad;

	d = Distance(lat1, lon1, lat2, lon2, true);
	base = acos((sin(lat2)-sin(lat1)*cos(d))/(sin(d)*cos(lat1)));

	if (sin(lon2-lon1) > 0.0)
		c_rad = base;
	else c_rad = (2.0*M_PI)-base;

	if (!asRadians)
		return GC_RADIANS_TO_DEGREES(c_rad);
	else return c_rad;
}


/* Calculate an intermediate point (latx/lonx) between two positions
given an interval distance between the two positions.

Parameter       Direction     Units
-----------------------------------
lat1               ->          rad
lon1               ->          rad
lat2               ->          rad
lon2               ->          rad
latx               <-          rad
lonx               <-          rad
interval_dist      ->          nm
total_dist         ->          nm
<return>           <-          bool
*/
bool GreatCircle::IntermediateByDistance(double lat1, double lon1,
										 double lat2, double lon2,
										 double *latx, double *lonx,
										 double interval_dist,
										 double total_dist)
{
	float fraction;
	bool ret;

	if (interval_dist > total_dist)
		return false;

	/* Calculate fraction of total distance */
	fraction = (float)(interval_dist/total_dist);

	/* Call other Intermediat* function with these inputs */
	ret = IntermediateByFraction(lat1, lon1, lat2, lon2,
		latx, lonx, total_dist,
		fraction);

	/* Return success/failure */
	return ret;
}


/* Calculate an intermediate point (latx/lonx) between two positions
given a fractional distance between the two positions.

Parameter       Direction     Units
-----------------------------------
lat1               ->          rad
lon1               ->          rad
lat2               ->          rad
lon2               ->          rad
latx               <-          rad
lonx               <-          rad
total_dist         ->          nm
fraction           ->          none
<return>           <-          bool
*/
bool GreatCircle::IntermediateByFraction(double lat1, double lon1,
										 double lat2, double lon2,
										 double *latx, double *lonx,
										 double total_dist,
										 float fraction)
{
	double tdist_conv;
	double sin_tdist;
	double cos_lat1;
	double cos_lat2;
	double A;
	double B;
	double x;
	double y;
	double z;

	/* Make sure we have total distance in radians */
	tdist_conv = GC_NM_TO_RADIANS(total_dist);

	/* Calculate some often reused values */
	sin_tdist = sin(tdist_conv);
	cos_lat1 = cos(lat1);
	cos_lat2 = cos(lat2);

	/* The actual calulation */
	A = sin((1.0-fraction)*tdist_conv)/sin_tdist;
	B = sin(fraction*tdist_conv)/sin_tdist;
	x = (A*cos_lat1*cos(lon1))+(B*cos_lat2*cos(lon2));
	y = (A*cos_lat1*sin(lon1))+(B*cos_lat2*sin(lon2));
	z = (A*sin(lat1))+(B*sin(lat2));

	*latx = atan2(z, sqrt((x*x)+(y*y)));
	*lonx = atan2(y, x);

	return true;
}


/* Converts the distance in radians (rad) to nautical miles (Nm).
By definition, one nautical mile subtends one minute
(1/60 degree) of arc.

Parameter       Direction     Units
-----------------------------------
rad                ->          rad
<return>           <-          nm
*/
inline double GreatCircle::RadiansToNm(double rad)
{
	return ((180.0*60.0)/M_PI)*rad;
}

/* Converts the distance in nautical miles (Nm) to radians (rad).
By definition, one nautical mile subtends one minute
(1/60 degree) of arc.

Parameter       Direction     Units
-----------------------------------
nm                 ->          nm
<return>           <-          rad
*/
inline double GreatCircle::NmToRadians(double nm)
{
	return (M_PI/(180.0*60.0))*nm;
}

/* Converts nautical miles (Nm) to meters (m).

Parameter       Direction     Units
-----------------------------------
nm                 ->          nm
<return>           <-          m
*/
double GreatCircle::NmToMeters(double nm)
{
	return nm*1852.0;
}

/* Converts meters (m) to nautical miles (Nm).

Parameter       Direction     Units
-----------------------------------
m                  ->          m
<return>           <-          Nm
*/
inline double GreatCircle::MetersToNm(double m)
{
	return m/1852.0;
}


/* Converts the angle in radians (rad) to degrees

Parameter       Direction     Units
-----------------------------------
rad                ->          rad
<return>           <-          deg
*/
inline double GreatCircle::RadiansToDegrees(double rad)
{
	return (180.0/M_PI)*rad;
}

/* Converts the angle in degrees (deg) to radians

Parameter       Direction     Units
-----------------------------------
deg                ->          deg
<return>           <-          rad
*/
double GreatCircle::DegreesToRadians(double deg)
{
	return (M_PI/180.0)*deg;
}
