#include "path.h"
#include "dted.h"
#include "site.h"
#include "coord.h"
#include "geod.h"
#include "exceptions.h"

/* This function implements the arc cosine function,
returning a value between 0 and M_PI*2. */
double Path::_arccos(double x, double y)
{
	double result = 0.0;

	if (y > 0.0)
		result = acos(x/y);

	if (y < 0.0)
		result = PI+acos(x/y);

	return result;
}



Path::Path()
{
	m_vCoords.clear();
	m_pSource = NULL;
	m_pDest = NULL;
	m_lObstruction = -1;
	m_iObstructionAlt = -6000;
}

Path::Path(Site *src, Site *dest, DTED *cell, bool interpalt)
{
	static float incarr[5] = { 0.07f, 0.02f, 0.007f, 0.002f, 0.0007f };
	float inc = incarr[cell->Level()];

	m_lObstruction = -1;
	m_iObstructionAlt = -6000;
	CreatePath(src, dest, cell, NULL, true, inc, interpalt);
}

Path::~Path()
{
	while (m_vCoords.size() != 0)
	{
		delete m_vCoords[0];
		m_vCoords.erase(m_vCoords.begin(), m_vCoords.begin()+1);
	}
}

Coord *Path::Location(int index)
{
	if (0 <= index && index < (int)m_vCoords.size())
		return m_vCoords[index];
	else return NULL;
}

Site *Path::LocationSite(int index)
{
	if (0 <= index && index < (int)m_vCoords.size())
	{
		Site *p = new Site("",
			m_vCoords[index]->Latitude,
			m_vCoords[index]->Longitude,
			m_vCoords[index]->Altitude,
			0.0);
		return p;
	}
	else return NULL;
}

int Path::AddPoint(double lat, double lon, double alt, double dis)
{
	Coord *c = new Coord(lat, lon, alt, dis);
	m_vCoords.push_back(c);
	return (int)m_vCoords.size()-1;
}

void Path::DelPoint(int index)
{
	if (0 <= index && index < (int)m_vCoords.size())
	{
		vector<Coord*>::iterator v;
		int i = 0;

		v = m_vCoords.begin();

		while (i < index)
			v++, i++;

		delete m_vCoords[index];
		m_vCoords.erase(v, v+1);
	}
	else
	{
		stringstream s("List index out of bounds: ");
		s << index;
		throw EListError(s.str());
	}
}

double Path::Latitude(int index)
{
	return _getcoordvalue(index, 0);
}

double Path::Longitude(int index)
{
	return _getcoordvalue(index, 1);
}

double Path::Altitude(int index)
{
	return _getcoordvalue(index, 2);
}

double Path::Distance(int index)
{
	return _getcoordvalue(index, 3);
}


/* This function generates a sequence of latitude and longitude positions
between a source location and a destination along a great circle path,
and stores elevation and distance information for points along that
path in the "path" structure for later use. */
bool Path::CreatePath(Site *src, Site *dest, DTED *cell, DTEDLookupCallback callBack,
					  bool include_lastpos, float inc, bool interpalt)
{
	double azimuth, distance, lat1, lon1, beta;
	double den, num, lat2, lon2, total_distance;
	double altstep = 0.0, diststep = 0.0;
	int x0, y0, x1, y1;
	Site tempsite;
	DTED *lookup = NULL;

	m_pSource = src;
	m_pDest = dest;

	x0=0;
	y0=0;

	/* latitude in radians */
	lat1 = src->Latitude()*DEG2RAD;
	/* longitude in radians */
	lon1 = src->Longitude()*DEG2RAD;
	/* azimuth in radians, math coordinate frame */
	azimuth = (360.0-src->Azimuth(*dest))*DEG2RAD;
	/* distance in nautical miles */
	total_distance = *src-*dest;

	/* interpolate altitudes, i.e., for flight path */
	if (interpalt == true)
	{
		altstep = dest->Altitude()-src->Altitude();
		diststep = total_distance/inc;
		if (diststep != 0.0)
			altstep /= diststep;
	}

	for (int i = 0; i < (int)m_vCoords.size(); i++)
		delete m_vCoords[i];
	m_vCoords.clear();

	/*cout<<"Total distance: "<<total_distance<<endl;*/
	/*printf("path\tdist\tbeta\t\tlat2\t\tnum\t\tden\n");*/
	for (distance = 0; distance <= total_distance; distance += inc)
	{
		/*beta = distance/3959.0;*/
		beta = distance/3436.029;
		lat2 = asin(sin(lat1)*cos(beta)+cos(azimuth)*sin(beta)*cos(lat1));
		num = cos(beta)-(sin(lat1)*sin(lat2));
		den = cos(lat1)*cos(lat2);

		/*printf("\t%3.2f\t%3.6f\t%3.6f\t%3.6f\t%3.6f\n",
		distance, beta, lat2, num, den);*/

		if (azimuth == 0.0 && (beta > (M_PI/2.0)-lat1))
		{
			lon2 = lon1+PI;
			/*printf("1a:\tazimuth = %3.3f, beta = %3.6f, pi/2 = %3.6f, lat1 = %3.6f\n",
			azimuth, beta, M_PI/2.0, lat1);*/
		}
		else if (azimuth == HALFPI && (beta > (M_PI/2.0)+lat1))
		{
			lon2 = lon1+PI;
			/*printf("2a:\tazimuth = %3.3f, beta = %3.6f, pi/2 = %3.6f, lat1 = %3.6f\n",
			azimuth, beta, M_PI/2.0, lat1);*/
		}
		else if (fabs(num/den) > 1.0)
		{
			lon2 = lon1;
			/*printf("3a:\tnum = %3.6f, den = %3.6f, fabs() = %3.6f\n",
			num, den, fabs(num/den));*/
		}
		else
		{
			if ((M_PI-azimuth) >= 0.0)
			{
				lon2 = lon1-_arccos(num, den);
				/*printf("4a:\tlon1 = %3.6f, num = %3.6f, den = %3.6f, arccos(num, den) = %3.6f, lon2 = %3.6f\n",
				lon1, num, num, _arccos(num, den), lon2);*/
			}
			else
			{
				lon2 = lon1+_arccos(num, den);
				/*printf("4b:\tlon1 = %3.6f, num = %3.6f, den = %3.6f, arccos(num, den) = %3.6f, lon2 = %3.6f\n",
				lon1, num, num, _arccos(num, den), lon2);*/
			}
		}

		/* This while() test broke the implementation for a while; going from 118W => 117W would
		produce longitudes of 252E. */
		/*while (lon2 < 0.0)
		lon2 += (M_PI*2.0);*/

		while (lon2 > (M_PI*2.0))
		{
			lon2 -= (M_PI*2.0);
			/*printf("6a:\tlon2 > 2M_PI: lon2 = %3.6f\n", lon2);*/
		}

		lat2 = lat2/DEG2RAD;
		lon2 = lon2/DEG2RAD;

		x1 = (int)(1199.0*(lat2-floor(lat2)));
		y1 = (int)(1199.0*(lon2-floor(lon2)));

		if (x1 != x0 || y1 != y0)
		{
			tempsite.Latitude(lat2);
			tempsite.Longitude(lon2);

			if (callBack != NULL)
				lookup = callBack(lat2, lon2);
			else lookup = cell;

			if (lookup == NULL && cell != NULL)
				lookup = cell;

			/*printf("lookup:\t\t\t\t%f\t%f\n", lat2, lon2);*/
			if (lookup != NULL)
			{
				double alt;

				if (interpalt)
					alt = src->Altitude()+((distance/inc)*altstep);
				else alt = lookup->AltitudeBySite(&tempsite);

				AddPoint(lat2, lon2, alt, distance);
			}
		}

		x0 = x1;
		y0 = y1;
	}

	if (include_lastpos == true)
	{
		/* Make sure exact dest point is recorded at path.length-1 */
		/*printf("Dest lookup: lat = %f, lon = %f, lookup = %p\n", dest->Latitude(), dest->Longitude(), lookup);*/
		if (callBack != NULL)
			lookup = callBack(dest->Latitude(), dest->Longitude());
		else lookup = cell;

		if (lookup == NULL && cell != NULL)
			lookup = cell;

		if (lookup != NULL)
		{
			double alt;

			if (interpalt)
				alt = dest->Altitude();
			else alt = lookup->AltitudeBySite(dest);

			AddPoint(dest->Latitude(), dest->Longitude(),
				alt, total_distance);
		}
	}

	return true;
}

bool Path::SaveToFile(char *fname)
{
	int i;
	ofstream op(fname);

	op<<"!path-count: "<<(int)m_vCoords.size()<<endl;
	op<<"!path-obstruction: "<<m_lObstruction<<endl<<endl;

	for (i = 0; i < (int)m_vCoords.size(); i++)
	{
		Coord *p;

		p = m_vCoords[i];

		op<<"!index: "<<i<<endl;
		op<<"!latitude: "<<geod::pos2str(p->Latitude, 0)<<endl;
		op<<"!longitude: "<<geod::pos2str(p->Longitude, 1)<<endl;
		op<<"!altitude: "<<(int)p->Altitude<<endl;
		op<<"!distance: "<<p->Distance<<endl<<endl;
		op<<endl;
	}
	op.close();

	return true;
}

bool Path::SaveToFileForGnuplot(char *fname, int pn, int sn)
{
	int i;
	double d;
	ofstream op(fname);

	for (i = 0; i < (int)m_vCoords.size(); i++)
	{
		Coord *p;

		p = m_vCoords[i];
		d = p->Distance;
		d *= (66.0/76.0);
		d *= 5280.0;
		d *= 12.0;
		d /= 39.0;

		op<<i<<"\t";
		op<<d<<"\t";
		/*op<<p->Latitude()<<"\t";
		op<<p->Longitude<<"\t";*/
		op<<(int)p->Altitude;

		if (m_lObstruction != -1 && i == m_lObstruction)
			op<<"\t"<<(int)m_iObstructionAlt<<"\t"<<(int)m_iObstructionAlt;
		if (i == 0 && m_pSource != NULL)
			op<<"\t"<<(int)m_pSource->Altitude();
		else if (i == (int)m_vCoords.size()-1 && m_pDest != NULL)
		{
			op<<"\t"<<(int)m_pDest->Altitude();

			if (m_lObstruction != -1)
				op<<"\t"<<(int)m_pDest->Altitude();
		}
		op<<endl;
	}
	/*if (m_lObstruction != -1)
	{
	op<<endl;
	op<<m_lObstruction<<"\t!\t0\n";
	op<<m_lObstruction<<"\t!\t";
	op<<(int)m_vCoords[m_lObstruction]->Altitude()<<endl;
	op<<m_lObstruction<<"\t!\t0\n";
	}*/
	op.close();

	ofstream oc((string(fname)+".ctrl").c_str());
	char buf[256];
	_snprintf(buf, 255, "plot \"%s\" using 1:3 title \"terrain%d.%d\" with " \
		      "lines, \"%s\" using 1:4 title \"LOS\" with lines, \"%s\" using " \
			  "1:5 title \"No LOS\" with lines", fname, pn, sn, fname, fname);
	oc<<buf<<endl;
	oc.close();

	return true;
}

double Path::_getcoordvalue(int index, int field)
{
	if (0 <= index && index < (int)m_vCoords.size())
	{
		switch (field)
		{
		case 0: return m_vCoords[index]->Latitude;
		case 1: return m_vCoords[index]->Longitude;
		case 2: return m_vCoords[index]->Altitude;
		case 3: return m_vCoords[index]->Distance;
		default: return m_vCoords[index]->Latitude;
		}
	}
	else
	{
		stringstream s("List index out of bounds: ");
		s << index;
		throw EListError(s.str());
	}
}

int Path::TravelLOS(Site *src, Site *dest, int &obsalt)
{
	int i;
	long index, pathlen;
	bool block;
	double distance, tx_alt, rx_alt;
	double xmtr_angle, test_angle, test_alt;

	index = -1;
	pathlen = Length();
	obsalt = -6000;
	tx_alt = src->Altitude()+src->Antenna();
	rx_alt = dest->Altitude()+dest->Antenna();
	distance = Distance(Length()-1);

	/* distance in [nm] to [m] */
	distance *= (66.0/76.0);
	distance *= 5280.0;
	distance *= 12.0;
	distance /= 39.0;

	xmtr_angle = atan2(rx_alt-tx_alt, distance);

	/* Determine the elevation angle of each discrete location
	along the path between the receiver and transmitter. */
	for (i = 1, block = false; i < pathlen && block == false; i++)
	{
		Coord *c;

		c = Location(i);

		test_alt = c->Altitude;
		distance = c->Distance;

		/* distance in [nm] to [m] */
		distance *= (66.0/76.0);
		distance *= 5280.0;
		distance *= 12.0;
		distance /= 39.0;

		/* Determine the elevation of the test point as seen from
		the location of the receiver */
		test_angle = atan2(test_alt-tx_alt, distance);

		if (xmtr_angle < test_angle)
		{
			block = true;
			index = i;

#if 0
			while (xmtr_angle < test_angle)
			{
				test_alt -= 1.0;
				test_angle = atan2(test_alt-tx_alt, distance);
			}
			obsalt = (int)test_alt;
#endif
			/*printf("\tadjusting alt = %3.2f, cos(angle) = %f, cos(xmtr) = %f\n",
			test_alt, test_angle, xmtr_angle);*/
		}
	}

	return index;
}
