#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <algorithm>

#include "radial.h"

void Radial::_calculate(struct viewpoint *vp, vector<Radial::point> &list)
{
  int i;
  float r, f, angle;
  Radial::point endpt, dpt, *pt;
  
  /* clear list */
  list.clear();

  /* return if NULL pointer */
  if (!vp)
    return;

  /* generate list of radials */
  for (angle = 0.0; angle < 360.0; angle += vp->interval)
    list.push_back(Radial::point(angle));

  /* iterate through list and calculate endpoints on the bounding box */
  for (i = 0; i < (int)list.size(); i++)
    {
      /* get a pointer to the next point */
      pt = &list[i];
      
      /* handle case of 0 degrees */
      if (pt->a.deg == 0)
	{
	  pt->fill(vp->site.left, vp->count.lat-vp->dim.height, 0.0);
	  continue;
	}
      
      /* handle case of 90 degrees */
      if (pt->a.deg == 90)
	{
	  pt->fill(vp->count.lon-vp->dim.width, vp->site.up, 90.0);
	  continue;
	}
      
      /* handle case of 180 degrees */
      if (pt->a.deg == 180)
	{
	  pt->fill(vp->site.left, 0, 180.0);
	  continue;
	}
      
      /* handle case of 270 degrees */
      if (pt->a.deg == 270)
	{
	  pt->fill(0, vp->site.up, 270.0);
	  continue;
	}

      /* choose an arbitrary radius */
      r = max(vp->count.lat, vp->count.lon)*4;

      /* calculate point at the end of the radius */
      endpt.x = vp->site.xi+(int)(r*cos(pt->a.rad));
      endpt.y = vp->site.yi+(int)(r*sin(pt->a.rad));

#ifdef _RADIAL_DEBUG
      /* print point */
      printf("pt[%3.2f] = (%d, %d), radius = %f\n", pt->a.deg, endpt.x, endpt.y, r);
#endif

      /* traverse path until a boundary is hit */
      for (f = 0.0; f < r; f += 0.5)
	{
	  float dx, dy;
	  
	  /* calculate new point */
	  dx = f*cos(pt->a.rad);
	  dy = f*sin(pt->a.rad);
	  dpt.x = vp->site.xi+(int)dx;
	  dpt.y = vp->site.yi+(int)dy;

	  /* continue if point is still in bounds */
	  if (0 < dpt.x && dpt.x < vp->count.lon-vp->dim.width &&
	      0 < dpt.y && dpt.y < vp->count.lat-vp->dim.height)
	    continue;

#ifdef _RADIAL_DEBUG
	  /* check for a top intersection */
	  if (dpt.y >= vp->count.lat-vp->dim.height)
	    printf("\t*** hit top at (%d, %d)\n", dpt.x, dpt.y);
	  
	  /* check for a bottom intersection */
	  else if (dpt.y == 0)
	    printf("\t*** hit bottom at (%d, %d)\n", dpt.x, dpt.y);

	  /* check for a left intersection */
	  else if (dpt.x == 0)
	    printf("\t*** hit left at (%d, %d)\n", dpt.x, dpt.y);

	  /* check for a right intersection */
	  else if (dpt.x >= vp->count.lon-vp->dim.width)
	    printf("\t*** hit right at (%d, %d)\n", dpt.x, dpt.y);
#endif

	  break;
	}
      
      /* update point */
      pt->x = dpt.x;
      pt->y = dpt.y;
      pt->a.deg = 90-pt->a.deg;
      while (pt->a.deg < 0.0)
	pt->a.deg += 360.0;
    }

  /* sort the vector by heading in degrees */
  std::sort(list.begin(), list.end());
}


Radial::point *Radial::Point(int index)
{
  /* check if indexed correctly */
  if (0 <= index && index < (int)m_vList.size())
    return &m_vList[index];

  /* find failed */
  return NULL;
}

Radial::point *Radial::Point(float bearing)
{
  vector<point>::iterator p;

  /* find entry */
  for (p = m_vList.begin(); p != m_vList.end(); p++)
    {
      if ((*p).a.deg == bearing)
	return &(*p);
    }

  /* find failed */
  return NULL;
}

void Radial::Create(int pov_x, int pov_y, float interval,
		    int height, int width, int latcnt, int loncnt)
{
  struct viewpoint pov;

  /* fill structure */
  pov.site.xi = pov_x;
  pov.site.yi = pov_y;
  pov.interval = interval;
  pov.dim.height = height;
  pov.dim.width = width;
  pov.count.lat = height*latcnt;
  pov.count.lon = width*loncnt;

  /* determine distances from boundaries to POV */
  pov.calc_bound_dist();
  
  /* calculate the radial values */
  _calculate(&pov, m_vList);
}

#ifdef _TEST_MAIN
int main(int argc, char *argv[])
{
  int i;
  Radial rad;
  Radial::point *p;
  FILE *fp;

  /* check parameters */
  if (argc != 2)
    {
      printf("Usage: %s interval\n", argv[0]);
      exit(1);
    }

  /* create and execute a new Radial object */
  rad.Create(600, 600, atof(argv[1]), 1, 1, 1201, 1201);
  
  /* display number of elements in the list */
  printf("Built %d locations\n", rad.Count());

  /* dump output to a file */
  fp = fopen("radial.txt", "w+");

  /* check for errors */
  if (!fp)
    return 1;

  fprintf(fp, "Angle\tX\tY\n");
  for (i = 0; i < (int)rad.Count(); i++)
    {
      /* get a pointer to the indicated index */
      p = rad.Point(i);
      
      fprintf(fp, "%3.1f\t%d\t%d\n", p->a.deg, p->x, p->y);
    }

  fclose(fp);
  
  /* return success */
  return 0;
}
#endif
