#include "site.h"
#include "coord.h"
#include "geod.h"
#include "greatcircle.h"

Site::Site()
{
	m_sName = "";
	m_dLatitude = 0;
	m_dLongitude = 0;
	m_dAltitude = 0;
	m_fAntenna = 0;
#ifdef _SITE_UTM
	m_lEasting = 0;
	m_lNorthing = 0;
	m_sZone = "";
	m_iDatum = 100;
#endif
	m_bValid = false;
	m_bActive = true;
}

Site::Site(char *name, double lat, double lon, double alt, float ant)
{
	m_sName = string(name);
	m_dLatitude = lat;
	m_dLongitude = lon;
	m_dAltitude = alt;
	m_fAntenna = ant;
#ifdef _SITE_UTM
	m_lEasting = 0;
	m_lNorthing = 0;
	m_sZone = "";
	m_iDatum = 100;
#endif
	m_bValid = true;
	m_bActive = true;
}

Site::Site(char *fname)
{
	ifstream *p;
	string line;
	size_t length;
	bool b[3] = {false, false, false};

	try 
	{
		p = new ifstream(fname);
	} 
	catch (...) 
	{
		cout<<"error opening site file: "<<fname<<endl; 
		delete this; 
		return; 
	}

	while (getline(*p, line, '\n'))
	{
		if (line.find("#", 0) == 0)
			continue;
		if (line.find("!name:", 0) != string::npos)
		{
			length = strlen("!name:");
			line = line.substr(length, line.length()-length);
			if (line.find(" ", 0) == 0)
				line = line.substr(1, line.length()-1);
			m_sName = line;
		}
		else if (line.find("!latitude:", 0) != string::npos)
		{
			length = strlen("!latitude:");
			line = line.substr(length, line.length()-length);
			m_dLatitude = geod::str2pos((char*)line.c_str());
			b[0] = true;
		}
		else if (line.find("!longitude:", 0) != string::npos)
		{
			length = strlen("!longitude:");
			line = line.substr(length, line.length()-length);
			m_dLongitude = geod::str2pos((char*)line.c_str());
			b[1] = true;
		}
		else if (line.find("!altitude:", 0) != string::npos)
		{
			length = strlen("!altitude:");
			line = line.substr(length, line.length()-length);
			stringstream v(line);

			v >> m_dAltitude;
			b[2] = true;
		}
		else if (line.find("!antenna:", 0) != string::npos)
		{
			length = strlen("!antenna:");
			line = line.substr(length, line.length()-length);
			stringstream v(line);

			v >> m_fAntenna;
		}
#ifdef _SITE_UTM
		else if (line.find("!easting:", 0) != string::npos)
		{
			length = strlen("!easting:");
			line = line.substr(length, line.length()-length);
			stringstream v(line);

			v >> m_lEasting;
		}
		else if (line.find("!northing:", 0) != string::npos)
		{
			length = strlen("!northing:");
			line = line.substr(length, line.length()-length);
			stringstream v(line);

			v >> m_lNorthing;
		}
		else if (line.find("!zone:", 0) != string::npos)
		{
			length = strlen("!zone:");
			line = line.substr(length, line.length()-length);
			m_sZone = line;
		}
		else if (line.find("!datum:", 0) != string::npos)
		{
			length = strlen("!datum:");
			line = line.substr(length, line.length()-length);
			stringstream v(line);

			v >> m_iDatum;
		}
#endif
		//else cout<<"found line: \""<<line<<"\"\n";
	}
	p->close();
	delete p;

	if (b[0] == false)
		cout<<"Need site latitude\n";
	if (b[1] == false)
		cout<<"Need site longitude\n";
	if (b[2] == false)
		cout<<"Need site altitude\n";

	m_bValid = (b[0] && b[1] && b[2]);
	m_bActive = true;
}


/* [ result = great circle distance in nautical miles between 'this' and 'c' ] */
double Site::operator-(Site &c)
{
	double lat1, lon1, lat2, lon2, distance;

	lat1 = m_dLatitude*DEG2RAD;
	lon1 = m_dLongitude*DEG2RAD;
	lat2 = c.m_dLatitude*DEG2RAD;
	lon2 = c.m_dLongitude*DEG2RAD;

	distance = GreatCircle::Distance(lat1, lon1, lat2, lon2, false);

	return distance;
}

/* [ result = great circle distance in nautical miles between 'this' and 'c' ] */
double Site::operator-(Coord &c)
{
	double lat1, lon1, lat2, lon2, distance;

	lat1 = m_dLatitude*DEG2RAD;
	lon1 = m_dLongitude*DEG2RAD;
	lat2 = c.Latitude*DEG2RAD;
	lon2 = c.Longitude*DEG2RAD;

	distance = GreatCircle::Distance(lat1, lon1, lat2, lon2, false);

	return distance;
}

bool Site::operator==(Site &c)
{
	bool b;

	b = (m_dLatitude == c.m_dLatitude &&
		m_dLongitude == c.m_dLongitude &&
		m_dAltitude == c.m_dAltitude &&
		m_fAntenna == c.m_fAntenna);

	return b;
}

/* [ result = azimuth in degrees from 'this' to 'dest' ] */
double Site::Azimuth(Site &dest)
{
	double dest_lat, dest_lon, src_lat, src_lon;
	double azimuth;

	dest_lat = dest.m_dLatitude*DEG2RAD;
	dest_lon = dest.m_dLongitude*DEG2RAD;
	src_lat = m_dLatitude*DEG2RAD;
	src_lon = m_dLongitude*DEG2RAD;

	azimuth = GreatCircle::Radial(src_lat, src_lon, dest_lat, dest_lon, false);

	return azimuth;
}

double Site::operator^(Site &c)
{
	return Elevation(c);
}

/* [ result = elevation in degrees from 'this' to 'dest' ] */
double Site::Elevation(Site &dest)
{
	/* This function returns the angle of elevation (in degrees)
	of the remote location as seen from the local site.
	A positive result represents an angle of elevation (uptilt),
	while a negative result represents an angle of depression
	(downtilt), as referenced to a normal to the center of
	the earth. */

	double a, b, dx;

	/* altitude [m] of 'dest' */
	a = dest.m_dAltitude+dest.m_fAntenna;

	/* altitude [m] of 'this' */
	b = m_dAltitude+m_fAntenna;

	/* distance [nm] from 'this' to 'dest' */
	dx = dest-*this;

	/* convert [nm] to [m] */
	dx *= (66.0/76.0);
	dx *= 5280.0;
	dx *= 12.0;
	dx /= 39.0;

	double dz = a-b;
	double d2 = dx;
	double d3_sq = (d2*d2)+(dz*dz);
	double d3 = sqrt(d3_sq);
	double theta = atan2(dz, d3) / DEG2RAD;

	return theta;
}

/* [ result = decimal representation of latitude string 'str' ] */
double Site::_str2lat(char *str, int len)
{
	/* [ c = last character in 'str' in lowercase ] */
	char c = tolower(str[len-1]);

	/* [ c == 'n' or c == 's' => result = 'str' as decimal latitude
	| true => result = -65535 ] */
	if (c == 'n' || c == 's')
	{
		// DTED style (DDDMMSSH)
		int deg, min, sec, dir;
		sscanf(str, "%3d%2d%2d%c", &deg, &min, &sec, &c);
		dir = (tolower(c) == 'n')?1:-1;
		return dir*((float)deg+((float)min/60.0)+((float)sec/3600.0));
	}
	else return -65535;
}

/* [ result = decimal representation of longitude string 'str' ] */
double Site::_str2lon(char *str, int len)
{
	/* [ c = last character in 'str' in lowercase ] */
	char c = tolower(str[len-1]);

	/* [ c == 'w' or c == 'e' => result = 'str' as decimal longitude
	| true => result = -65535 ] */
	if (c == 'w' || c == 'e')
	{
		// DTED style (DDDMMSSH)
		int deg, min, sec, dir;
		sscanf(str, "%3d%2d%2d%c", &deg, &min, &sec, &c);
		dir = (tolower(c) == 'e')?1:-1;
		return dir*((float)deg+((float)min/60.0)+((float)sec/3600.0));
	}
	else return -65535;
}

void Site::Print(ostream *fp)
{
	ostream *p;

	if (fp == NULL)
		p = &cout;
	else p = fp;

	*p<<"Name:      "<<m_sName<<endl;
	*p<<"Latitude:  "<<geod::pos2str(m_dLatitude, 0)<<endl;
	*p<<"Longitude: "<<geod::pos2str(m_dLongitude, 1)<<endl;
	*p<<"Altitude:  "<<m_dAltitude<<endl;
	*p<<"Antenna:   "<<m_fAntenna<<endl;
	*p<<endl;
}

void Site::SetLatitude(int deg, int min, float sec, char dir)
{
	double v = (double)deg;

	v += ((double)min/60.0);
	v += sec/3600.0;

	if (dir == 'S' || dir == 's')
		v *= -1.0;

	if (_valid_lat(v))
		m_dLatitude = v;
}

void Site::SetLongitude(int deg, int min, float sec, char dir)
{
	double v = (double)deg;

	v += ((double)min/60.0);
	v += sec/3600.0;

	if (dir == 'W' || dir == 'w')
		v *= -1.0;

	if (_valid_lon(v))
		m_dLongitude = v;
}


#ifdef _SITE_MAIN
int main(int argc, char *argv[])
{
	Site *p, *q;
	double distance, bearing;

	if (argc != 3)
	{
		printf("Usage: %s <site1> <site2>\n", argv[0]);
		return 1;
	}

	p = new Site(argv[1]);
	q = new Site(argv[2]);

	distance = *q-*p;
	bearing = p->azimuth(*q);

	p->Print();
	q->Print();

	printf("\nDistance: %f\n", distance);
	printf("Bearing: %f\n", bearing);

	delete p;
	delete q;
}
#endif
