#ifndef _analysisthread_h
#define _analysisthread_h

#include "wndthreadbase.h"

#include <map>
#include <vector>
#include <string>
#include <algorithm>

#include "resource.h"
#include "optlosDlg.h"

#include "plotter.h"
#include "collection.h"
#include "graphdtedgrid_win32.h"
#include "path.h"
#include "dtedgrid.h"
#include "dted.h"
#include "site.h"
#include "coord.h"

class AnalysisThread: public WndThreadBase<COptLosDlg>
{
protected:
	struct _site_index
	{
		int index;
		double dist;
		Site *ptr;

		bool operator <(struct _site_index &c)
		{
			return (dist < c.dist);
		}
	};

	void map_color(int i_emt, unsigned short *r, unsigned short *g,
		unsigned short *b)
	{
		switch (i_emt)
		{
		case 0:		/* white */
			*r = 0xffff, *g = 0xffff, *b = 0xffff;
			break;
		case 1:		/* blue */
			*r = 0, *g = 0, *b = 0xffff;
			break;
		case 2:		/* cyan */
			*r = 0xffff, *g = 0xffff, *b = 0;
			break;
		case 3:
			*r = 0, *g = 0xffff, *b = 0xffff;
			break;
		case 4:
			*r = 0xffff, *g = 0, *b = 0xffff;
			break;
		}
	}

public:
	AnalysisThread(CWnd *pParent, Collection *flightcol, 
				   Collection *emittercol, CString &dtedpath, 
				   float stepsize, std::map<std::string,bool> &flags) 
				   : WndThreadBase<COptLosDlg>(pParent, true),
			m_pGrid(NULL),
			m_pGraph(NULL),
			m_pFlightCol(flightcol),
			m_pEmitterCol(emittercol),
			m_sDtedPath(dtedpath),
			m_fStepSize(stepsize),
			m_mControlFlags(flags)
	{
		int i;
		CString str;

		/* determine bounds of the flight collection */
		m_pFlightCol->CoordBounds(m_pEmitterCol, &m_dBoundsXMin, &m_dBoundsYMin, 
								  &m_dBoundsXMax, &m_dBoundsYMax, false);

		/* load the needed DTED cells */
		m_pGrid = DTEDGrid::Load(m_dBoundsXMin, m_dBoundsYMin, m_dBoundsXMax, 
								 m_dBoundsYMax, (char*)m_sDtedPath.GetString(), true);
		if (m_pGrid == NULL)
		{
			str = "Failed to load DTED data (incorrect path?)";
			privThreadStatusCallback(str);

			Terminate();
			throw;
		}

		/* set status callback for DTEDGrid object */
	    m_pGrid->SetStatusCallback(GridStatusCB, (void*)this);

		m_pGraph = new GraphDtedGrid(GraphDtedGrid::gtBMP, m_pGrid);
		if (!m_pGraph->IsOpen())
		{
			privThreadStatusCallback("Graph not opened!");

			Terminate();
			throw;
		}

		/* fill emitter index list */
		for (i = 0; i < (int)m_pEmitterCol->SiteCount(); i++)
		{
			struct _site_index p;

			p.index = i;
			p.ptr = m_pEmitterCol->Sites(i);
			p.dist = 0;

			m_vEmitterIndices.push_back(p);
		}
	}

	BMPPlotter *GetPlotter()
	{
		if (m_pGraph == NULL)
			return NULL;
		return (BMPPlotter*)m_pGraph->Handle();
	}

	void Execute()
	{
		bool b;
		int i, j, fltpathmax, emtpathmax;
		Path ray;
		char buf[64] = "";
		int obsalt, obsidx;
		int max_width = -1, max_height = -1;
		int line_width = 2;
		CString str;
		unsigned short color[3];
		bool mincoverage;
		struct _nv_pos notvis;

		color[0] = 0;
		color[1] = 0;
		color[2] = 0;

		/* set local flags so lookups need not occur */
		mincoverage = m_mControlFlags["MinCoverage"];

		m_pFlightCol->SetGrid(m_pGrid);
		b = m_pFlightCol->GeneratePath(m_fStepSize, true);
		if (b == true)
			str.Format("Generated %d paths along flight path", m_pFlightCol->PathCount());
		else str = "Failed to generate paths";
		privThreadStatusCallback(str);

		privThreadStatusCallback("Drawing plot...");

		if (m_mControlFlags["RenderDted"] == true)
			m_pGraph->Generate();

		privThreadStatusCallback("Drawing emitter sites");
		for (int k = 0; k < (int)m_pEmitterCol->SiteCount(); k++)
		{
			map_color(k, &color[0], &color[1], &color[2]);
			m_pGraph->DrawSite(m_pEmitterCol->Sites(k), color, 35.0, true);
		}

		privThreadStatusCallback("Drawing flight paths");
		for (int k = 0; k < (int)m_pFlightCol->PathCount(); k++)
		{
			unsigned short color[3] = { 0, 255, 255};

			m_pGraph->DrawPath(m_pFlightCol->Paths(k), color, 15);
		}

		if (m_mControlFlags["Analyze"] == false)
		{
			if (m_pGraph)
			{
				m_pGraph->Close();
				delete m_pGraph;
			}

			goto _delete_grid;
		}

		for (i = 0; i < (int)m_pEmitterCol->SiteCount(); i++)
		{
			Site *pov;
			
			pov = m_pEmitterCol->Sites(i);

			str.Format("Emitter: %s, %s", pov->Name().c_str(), (pov->Active())?"active":"inactive");
			privThreadStatusCallback(str);
		}

		/* run flight paths to backtrace LOS to site pov */
		fltpathmax = m_pFlightCol->PathCount();

		/* run emitter paths to backtrace LOS */
		emtpathmax = m_pEmitterCol->SiteCount();
		str.Format("Emitter paths: %d", emtpathmax);
		privThreadStatusCallback(str);

		/* for each step in the path, check a ray to the static site */
		str = "Checking for line-of-sight obstructions...";
		privThreadStatusCallback(str);

		m_pGraph->Handle()->pencolor(0, 0xffff, 0);

		/* setup progress bar */
		m_pParentClass->prgAnalysis.SetRange32(0, fltpathmax);
		m_pParentClass->prgAnalysis.SetPos(0);

		for (i = 0; i < fltpathmax && !m_bTerminated; i++)
		{
			Path *p = m_pFlightCol->Paths(i);
			Site *s = m_pFlightCol->Sites(i);

			//if (i % 100 == 0)
			{
				m_pParentClass->prgAnalysis.SetPos(i);

				str.Format("%3.1f%%", ((float)i/(float)fltpathmax)*100.f);
				m_pParentClass->lblProgress.SetWindowText(str);
			}

			/* setup path progress bar */
			m_pParentClass->prgAnalysis2.SetRange32(0, (int)p->Length());
			m_pParentClass->prgAnalysis2.SetPos(0);

			int r = (int)p->Length();

			for (j = 0; j < (int)p->Length() && !m_bTerminated; j++)
			{
				Coord *c1 = p->Location(j);

				if (j % 10 == 0)
				{
					m_pParentClass->prgAnalysis2.SetPos(j);

					str.Format("%3.1f%%", ((float)j/(float)(int)p->Length())*100.f);
					m_pParentClass->lblProgress2.SetWindowText(str);
				}

				if (j % 100 == 0)
				{
					SortEmittersByDistance(c1);
				}

				//printf("path length: %d\n", (int)p->Length());
				for (int i_emt = 0; i_emt < emtpathmax && !m_bTerminated; i_emt++)
				{
					int sel_emit_index;
					Site *pov;
					
					sel_emit_index = m_vEmitterIndices[i_emt].index;
					pov = m_vEmitterIndices[i_emt].ptr;
					//pov = m_pEmitterCol->Sites(i_emt);

					if (pov->Active() == false)
					{
						cout<<"Site \""<<pov->Name()<<"\" is inactive\n";
						continue;
					}

					Site dest1((char*)buf, 
						c1->Latitude,
						c1->Longitude,
						s->Altitude(), 0);

					/* create path between the static point 
					and the site destination */
					ray.CreatePath(pov, &dest1, NULL,
						DTEDGrid::_dtedlookupfunc, 
						true, m_fStepSize, false);

					/* travel the path looking for obstructions */
					obsidx = ray.TravelLOS(pov, &dest1, obsalt);

					if (obsidx == -1)
					{
						if (mincoverage == false)
						{
							map_color(i_emt, &color[0], &color[1], &color[2]);
						}
						else
						{
							color[0] = 0;
							color[1] = 0xffff;
							color[2] = 0;
						}

						m_pGraph->DrawPoint(&dest1, color, 15.0f);

						if (mincoverage == true)
						{
							//printf("i_emt = %d, skipping\n", i_emt);
							break;
						}
					}
					else
					{
						if (i_emt == 0)
						{
							color[0] = 0xffff;
							color[1] = 0;
							color[2] = 0;
							m_pGraph->DrawPoint(&dest1, color, 15.0f);
						}

						notvis.lat = dest1.Latitude();
						notvis.lon = dest1.Longitude();
						m_vNotVisible.push_back(notvis);
					}
				}
			}      

#if 0
			if (i == 10)
				break;
#endif
		}

		/* redraw bad (not visible) positions */
		if (m_vNotVisible.size() > 0)
		{
			/* post status */
			str.Format("Redrawing %d bad positions", m_vNotVisible.size());
			privThreadStatusCallback(str);

			Site nv;
			color[0] = 255;
			color[1] = 0;
			color[2] = 0;

			for (i = 0; i < (int)m_vNotVisible.size(); i++)
			{
				nv.Latitude(m_vNotVisible[i].lat);
				nv.Longitude(m_vNotVisible[i].lon);
				m_pGraph->DrawPoint(&nv, color, 15.0f);
			}
		}

		/* show terminated message if Terminated() */
		if (m_bTerminated)
			privThreadStatusCallback("Analysis terminated");

		/* update progress controls */
		m_pParentClass->prgAnalysis.SetPos(fltpathmax);
		str.Format("%3.1f%%", ((float)fltpathmax/(float)fltpathmax)*100.f);
		m_pParentClass->lblProgress.SetWindowText(str);

		/* close graph and write to file */
		m_pGraph->Close();

		/* free graph */
		delete m_pGraph;

_delete_grid:
		/* free the DTED grid */
		delete m_pGrid;

		/* set Terminated flag */
		m_bAmTerminated = true;
	}

    static void GridStatusCB(void *sender, const char *str)
	{
		AnalysisThread *thog;

		thog = (AnalysisThread*)sender;
		thog->privThreadStatusCallback(str);
	}

	void privThreadStatusCallback(const char *str)
	{
		int count;

		count = m_pParentClass->lstStatus.GetCount();
		m_pParentClass->lstStatus.AddString(str);
		m_pParentClass->lstStatus.SetCurSel(count);
	}

	void privThreadStatusCallback(CString &str)
	{
		int count;

		count = m_pParentClass->lstStatus.GetCount();
		m_pParentClass->lstStatus.AddString(str);
		m_pParentClass->lstStatus.SetCurSel(count);
	}

private:
	void SortEmittersByDistance(Coord *dest)
	{
		int i, r;
		std::vector<struct _site_index>::iterator p;

		/* calculate distances from dest to each emitter */
		for (p = m_vEmitterIndices.begin(); p != m_vEmitterIndices.end(); p++)
			(*p).dist = (*(*p).ptr)-(*dest);

		/* sort vector by distance */
		std::sort(m_vEmitterIndices.begin(), m_vEmitterIndices.end());

		for (i = 0; i < (int)m_vEmitterIndices.size(); i++)
		{
			struct _site_index *t;

			t = &m_vEmitterIndices[i];
			r = 1;
		}
	}

protected:
	struct _nv_pos
	{
		double lat;
		double lon;
	};

	DTEDGrid *m_pGrid;
	GraphDtedGrid *m_pGraph;
	Collection *m_pFlightCol;
	Collection *m_pEmitterCol;
	std::vector<struct _site_index> m_vEmitterIndices;
	std::vector<struct _nv_pos> m_vNotVisible;
	CString m_sDtedPath;
	float m_fStepSize;
	std::map<std::string,bool> m_mControlFlags;
	double m_dBoundsXMin;
	double m_dBoundsXMax;
	double m_dBoundsYMin;
	double m_dBoundsYMax;
};

#endif
