#include "graphdted_win32.h"
#include "dted.h"
#include "path.h"
#include "coord.h"
#include "site.h"

#include "plotter.h"

//#define _COLOR_CHANNEL_IS_USHORT

GraphDted::GraphDted(int type, DTED *d) : GraphBase(type),
	m_pCell(d)
{
	Allocate(-1, -1);
}

GraphDted::~GraphDted()
{
	/* do not free the DTED pointer, as we don't own it */
}

bool GraphDted::Allocate(int maxwidth, int maxheight)
{
	int xsz = WINDOW_SIZE, ysz = WINDOW_SIZE;
	char buf[64];
	float xn, yn;
	int x_max, y_max;

	if (m_pCell == NULL)
		return false;

	x_max = m_pCell->LongitudeCount();
	y_max = m_pCell->LatitudeCount();

	m_pCell->CalcWindowSize(xsz, ysz, xn, yn);

	_snprintf(buf, 64-1, "%dx%d", xsz, ysz);
	Plotter::parampl("BITMAPSIZE", buf);

	switch (m_iPlotterType)
	{
	case gtX: 
		m_pHandle.x = new XPlotter(stdin);
		break;

	case gtPNG:
		m_pOutputFile = fopen("plot.png", "w+");
		m_pHandle.png = new PNGPlotter(m_pOutputFile);
		break;

	case gtBMP:
		m_pOutputFile = fopen("plot.bmp", "w+");
		m_pHandle.bmp = new BMPPlotter(m_pOutputFile);
		break;

	default:
		cout<<"GraphDted::Allocate(): Invalid plotter type specified: "<<m_iPlotterType<<endl;
		return false;
	}

	if (m_pHandle.p->openpl() < 0)
	{
		m_bOpen = false;
		return false;
	}
	m_bOpen = true;

	if (xn == 1.0 && yn == 1.0)
		m_pHandle.p->fspace (0.0f, 0.0f, (float)x_max, (float)y_max);
	else m_pHandle.p->fspace (0.0f, 0.0f, x_max*xn, y_max*yn);

	return true;
}

bool GraphDted::Generate()
{
	if (m_pCell == NULL)
		return false;

	m_iMinAltitude = m_pCell->MinAltitude();
	m_iMaxAltitude = m_pCell->MaxAltitude();

	return DrawCell(m_pCell, 0, 0);
}

bool GraphDted::DrawPath(Path *path, unsigned short col[3], int size)
{
	if (m_pCell == NULL || path == NULL || m_pHandle.p == NULL || m_bOpen == false)
		return false;

#ifdef _NO_PLOTXY
	return true;
#else
	int i;
	Coord *p1;
	Coord *p2;

	m_pHandle.p->pencolor(col[0], col[1], col[2]);
	m_pHandle.p->linewidth(size);

	for (i = 0; i < (int)path->Length()-1; i++)
	{
		ipoint2_t ip1 = {0, 0}, ip2 = {0, 0};

		p1 = path->Location(i);
		p2 = path->Location(i+1);

		m_pCell->MapCoordToPost(p1->Latitude, p1->Longitude, ip1.x, ip1.y);
		m_pCell->MapCoordToPost(p2->Latitude, p2->Longitude, ip2.x, ip2.y);

		m_pHandle.p->line(ip1.x, ip1.y, ip2.x, ip2.y);
	}

	return true;
#endif
}

void GraphDted::DrawSite(Site *p, unsigned short col[3], float size, bool showlabel)
{
#ifdef _NO_PLOTXY
	return;
#else
	ipoint2_t pt;

	if (m_pCell == NULL || m_pHandle.p == NULL || p == NULL)
		return;

	m_pHandle.p->pencolor(col[0], col[1], col[2]);
	m_pHandle.p->fillcolor(col[0], col[1], col[2]);

	m_pCell->MapCoordToGraphic(p->Latitude(), p->Longitude(), pt.x, pt.y);

	if (pt.x == -1 && pt.y == -1)
	{
		printf("GraphDted::DrawSite::map error\n");
		p->Print();
	}
	else 
	{
		/*printf("GraphDted::DrawSite::x = %d, y = %d\n", x, y);*/
		m_pHandle.p->marker(pt.x, pt.y, 2, (int)size);

		if (showlabel)
		{
			/* Show label */
			m_pHandle.p->alabel('c', 'b', (char*)p->Name().c_str());
		}
	}
#endif
}

void GraphDted::DrawPoint(Site *p, unsigned short col[3], float size)
{
#ifdef _NO_PLOTXY
	return;
#else
	ipoint2_t pt;

	if (m_pCell == NULL || m_pHandle.p == NULL || p == NULL)
	{
		printf("DrawPoint() error: p = %p, handle = %p, m_pCell = %p\n",
			p, m_pHandle.p, m_pCell);
		return;
	}

	m_pHandle.p->pencolor(col[0], col[1], col[2]);
	m_pHandle.p->fillcolor(col[0], col[1], col[2]);

	m_pCell->MapCoordToGraphic(p->Latitude(), p->Longitude(), pt.x, pt.y);

	if (pt.x == -1 && pt.y == -1 && 0)
	{
		printf("GraphDted::DrawPoint::map error\n");
		p->Print();
		return;
	}
	m_pHandle.p->point(pt.x, pt.y);
#endif
}

/* Takes 2 sites (i.e., points), calculates the slope between the two,
in order to calculate a point perpendicular to the resulting line
segment.  The point is assumed to be perpendicular to the first site,
on the left side. */
void GraphDted::DrawParallelPoint(Site *p1, Site *p2, unsigned short col[3], float size, int level)
{
#ifdef _NO_PLOTXY
	return;
#else
	ipoint2_t pt1, pt2;
	fpoint2_t npt;

	if (m_pCell == NULL || p1 == NULL || p2 == NULL || m_pHandle.p == NULL)
		return;

	m_pHandle.p->pencolor(col[0], col[1], col[2]);
	m_pHandle.p->fillcolor(col[0], col[1], col[2]);

	m_pCell->MapCoordToGraphic(p1->Latitude(), p1->Longitude(), pt1.x, pt1.y);
	m_pCell->MapCoordToGraphic(p2->Latitude(), p2->Longitude(), pt2.x, pt2.y);

	if ((pt1.x == -1 && pt1.y == -1) || (pt2.x == -1 && pt2.y == -1))
	{
		printf("GraphDted::DrawSite::map error\n");
		return;
	}

	{
		float dx, dy;
		float r1, r2;
		double theta1, theta2;

		dx = (float)(pt2.x-pt1.x);
		dy = (float)(pt2.y-pt1.y);

		/* convert p1 -> p2 to polar */
		/* distance from p1 to p2 */
		r1 = sqrt((dx*dx)+(dy*dy));
		/* theta1 is in radians */
		theta1 = atan2(dy, dx);

		/* rotate theta1 90deg right (clockwise) */
		theta2 = theta1+(M_PI/4.0);

		/* distance to next point is level*10 */
		r2 = (level+1)*5.0f;
		/* convert p1 -> p3 to cartesian */
		npt.y = (float)(r2*sin(theta2));
		npt.x = (float)(r2*cos(theta2));
		/* npt.x and npt.y are offsets; apply to p1 */

		/*printf("r1 = %3.2f, theta1 = %3.6f,\tr2 = %3.2f, theta2 = %3.6f,\tnpt.x = %3.2f, npt" \
		".y = %3.2f\n", r1, theta1, r2, theta2, npt.x, npt.y);*/
		npt.x += (float)pt1.x;
		npt.y += (float)pt1.y;

		//cout<<"npt.x: "<<npt.x<<", npt.y: "<<npt.y<<endl;
	}

	m_pHandle.p->fpoint(npt.x, npt.y);
#endif
}

void GraphDted::DrawParallelLine(Site *p1, Site *p2, unsigned short col[3], float size, int level)
{
#ifdef _NO_PLOTXY
	return;
#else
	ipoint2_t pt1, pt2;
	fpoint2_t npt1, npt2;

	if (m_pCell == NULL || p1 == NULL || p2 == NULL || m_pHandle.p == NULL)
		return;

	m_pHandle.p->pencolor(col[0], col[1], col[2]);
	m_pHandle.p->fillcolor(col[0], col[1], col[2]);

	m_pCell->MapCoordToGraphic(p1->Latitude(), p1->Longitude(), pt1.x, pt1.y);
	m_pCell->MapCoordToGraphic(p2->Latitude(), p2->Longitude(), pt2.x, pt2.y);

	if ((pt1.x == -1 && pt1.y == -1) || (pt2.x == -1 && pt2.y == -1))
	{
		printf("GraphDted::DrawSite::map error\n");
		return;
	}

	{
		float dx, dy;
		float r1, r2;
		double theta1, theta2;

		dx = (float)(pt2.x-pt1.x);
		dy = (float)(pt2.y-pt1.y);

		/* convert p1 -> p2 to polar */
		/* distance from p1 to p2 */
		r1 = sqrt((dx*dx)+(dy*dy));
		/* theta1 is in radians */
		theta1 = atan2(dy, dx);

		/* rotate theta1 90deg right (clockwise) */
		theta2 = theta1+(M_PI/4.0f);

		/* distance to next point is level*10 */
		r2 = (level+1)*size;
		/* convert p1 -> p3 to cartesian */
		dy = (float)(r2*sin(theta2));
		dx = (float)(r2*cos(theta2));
		/* dx and dy are offsets; apply to p1 and p2 */

		/*printf("r1 = %3.2f, theta1 = %3.6f,\tr2 = %3.2f, theta2 = %3.6f,\tnx = %3.2f, ny = %3.2f\n",
		r1, theta1, r2, theta2, nx, ny);*/
		npt1.x = dx+(float)pt1.x;
		npt1.y = dy+(float)pt1.y;
		npt2.x = dx+(float)pt2.x;
		npt2.y = dy+(float)pt2.y;

		//cout<<"nx: "<<nx<<", ny: "<<ny<<endl;
	}

	m_pHandle.p->fline(npt1.x, npt1.y, npt2.x, npt2.y);
#endif
}
